"""
    OptRecsUtils.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def debug_plot_opt_recs_impl(ax, x, y, opt_recs, color=None, spline=None):
    for k, rec in enumerate(opt_recs):
        func = rec.evaluator
        peak = rec.peak
        ax.plot(x, y, color="orange")
        ax.plot(x, func(x), ":", color=color)
        if spline is not None:
            px = peak.top_x
            py = spline(px)
            ax.plot(px, py, "o", color="yellow")

def debug_plot_opt_recs(ecurve, opt_recs, eval_x=None, title=None):

    x = ecurve.x
    y = ecurve.y
    max_y = ecurve.max_y
    if eval_x is None:
        eval_x = x

    with plt.Dp():
        if title is None:
            title = "OptRecsDebug"
        fig, ax = plt.subplots()
        ax.set_title(title)
        debug_plot_opt_recs_impl(ax, eval_x, y, opt_recs, spline=ecurve.spline)
        print("ratio=",y/max_y)
        fig.tight_layout()
        plt.show()

def compute_area_proportions(x, opt_recs):
    areas = []
    for k, rec in enumerate(opt_recs):
        func = rec.evaluator
        areas.append(np.sum(func(x)))
    areas = np.array(areas)
    return areas/np.sum(areas)

def eoii_correct_opt_recs(frame, opt_recs_, debug=False):
    if True:
        from importlib import reload
        import LRF.EoiiCorrector
        reload(LRF.EoiiCorrector)
        import DecompUtils
        reload(DecompUtils)

    from LRF.EoiiCorrector import EoiiCorrector
    from molass_legacy.DataStructure.AnalysisRangeInfo import convert_to_paired_ranges
    from DecompUtils import make_range_info_impl

    if debug:
        ecurve = frame.dialog.sd.xray_curve
        debug_plot_opt_recs(ecurve, opt_recs_, title="eoii_correct_opt_recs entry")

    """
        this is just a temporary implementation.
        make ranges for EoiiCorrector which only needs paird_range.top_x used in get_pno_map_impl()
        task: simplify this procudure to get ret[0], i.e., cnv_ranges in PreviewData
    """
    temp_paired_ranges = make_range_info_impl(None, frame.control_info, frame.specpanel_list, no_elm_recs=True)
    ret = convert_to_paired_ranges(temp_paired_ranges)

    ec = EoiiCorrector(frame.dialog.sd, ret[0])

    params_list = []
    for k, rec in enumerate(opt_recs_):
        values = rec.evaluator.get_all_param_values()
        if debug:
            print([k], values)
        params_list.append(values)

    func = opt_recs_[0].evaluator.get_func()  # egha or emga
    corrected, new_params_list = ec.correct_params_list(func, params_list, debug=debug)

    if corrected:
        for rec, new_params in zip(opt_recs_, new_params_list):
            rec.evaluator.set_new_params(new_params)        # this will induce a bug. do away with lmfit.

    if debug:
        ecurve = frame.dialog.sd.xray_curve
        debug_plot_opt_recs(ecurve, opt_recs_, title="eoii_correct_opt_recs result")



"""
    ElutionCurve.py

        recognition of peaks

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import copy
import logging
import numpy as np
from scipy.interpolate import LSQUnivariateSpline
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Elution.CurveUtils import proof_plot, simple_plot
from .CurveFeatures import CurveFeatures
from .OptimalSmoothing import OptimalSmoothing
from molass_legacy.DataStructure.PeakInfo import PeakInfo
from molass_legacy.DataStructure.AnalysisRangeInfo import PairedRange
from molass_legacy.Irregular.Preprocess import correct_flat_negative_baseline

DEBUG   = False

TOP_X_ALLOWANCE = 5
NUM_POINTS_PER_KNOT     = 10
MIN_NUM_KNOTS           = 50    # > 40 for pH6
MAX_NUM_KNOTS           = 100
NUM_KNOTS_MID_QUALITY   = 40
QRATIO_LIMIT            = 0.4   # > 0.273 for pH6, > 0.32 for 20170426
DETERMINE_KNOTS_ONCE    = False # must be False for Factin
NUM_KONTS_LOW_QUALITY   = 20    # < 25 for BL-10C/Ald
PEAK_INFO_FIND_RATIO    = 0.2
ANALYSIS_RANGE_RATIO    = 0.5
BETTER_RANGE_RATIO      = 0.1
SUFFICIENT_PEAK_H_RATIO = 0.3
MININUM_PEAK_WIDTH      = 10
MININUM_PEAK_W_RATIO    = 0.05
NEIGHBOR_IGNORE_RATIO   = 0.7
LOW_PEAK_CHECK_RATIO    = 0.7   # > 0.3 for 20170304_UV, > 0.59 for Factin
LOW_PEAK_CHECK_LIMIT    = 0.2   #
LOW_PEAK_RAISE_RATIO    = 1.02  # to allow the first low peak in Sugiyama
BETER_SPLINE_LIM_PNTS   = 20
BETER_SPLINE_LIM_WIDTH  = 0.5
MODEST_NOISE_LEVEL_LIM  = 0.01
FOOT_LEVEL_RATIO        = 0.1
FOOT_ADD_WIDTH          = 200
USE_HALF_MAXIMUM_POINTS = False
MAJOR_PEAK_RATIO        = 0.1
SEDIMENTATION_LIMIT = 0.03
END_DIFF_WIDTH = 50
END_DIFF_MIN_HEIGHT = SEDIMENTATION_LIMIT/2
RELIABLE_SIGMA_RATIO = 2
RELIABLE_ALPHA = 0.03
USE_RELIABLE_ALPHA = True
MIN_HALF_WIDTH = 5

class ElutionCurve:
    def __init__( self, y,
                    x=None,
                    max_y=None,
                    j0=0,
                    microfluidic=None,
                    unify_close_peaks=True,     # need be True for Factin
                    low_quality=False,
                    min_y_ratio=None,
                    debug_plot=False,
                    debug_smooth=False,
                    add_features=True,
                    cover_curve=None,
                    orig_top_x=None,
                    delay_emg_peaks=False,      # True for residual curves in ElutionDecomposer
                    possiblly_peakless=False ):

        self.logger = logging.getLogger( __name__ )
        if x is None:
            x = np.arange( len(y) )
        self.x  = x
        self.y  = correct_flat_negative_baseline(y)
        self.y_orig = y     # for backward compatibility
        m = np.argmax(y)
        self.max_x = x[m]   # task: rename to a better name
        self.max_y = y[m]
        self.min_peak_width = len(self.x) * MININUM_PEAK_W_RATIO
        self.min_y = np.percentile(y, 5)
        self.height = self.max_y - self.min_y
        self.j0 = j0
        self.mean = None
        self.variance = None
        self.peak_slice = slice(0, len(x))      # comptibility with PeakCurve
        self.peak_region_sigma = None
        self.end_slices = None

        self.low_quality = low_quality
        self.debug_plot = debug_plot

        if microfluidic is None:
            microfluidic = get_setting('use_mtd_conc')
        self.microfluidic = microfluidic

        self.smoothing = OptimalSmoothing(x, y, self.height, self.min_y, min_y_ratio=min_y_ratio, orig_top_x=orig_top_x)
        self.smoothing.compute_optimal_curves(microfluidic, debug_smooth)

        self.y_for_spline = self.smoothing.spline_y
        self.feature_y = self.smoothing.feature_y
        self.spline = self.smoothing.spline
        self.sy = self.spline(x)

        self.d1 = self.smoothing.d1
        self.d1y = self.d1(x)
        self.d2 = self.smoothing.d2
        self.d2y = self.d2(x)
        self.d2_roots = self.smoothing.d2_roots

        self.peak_top_x = self.get_significant_peaks( low_quality=low_quality, unify_close_peaks=unify_close_peaks )
        if len(self.peak_top_x ) == 0:
            raise RuntimeError("No peak_top_x")

        self.set_peak_dependent_info()
        self.emg_peaks = None
        # assure consistency between self.emg_peaks and peak_info
        self.possiblly_peakless = possiblly_peakless
        if delay_emg_peaks:
            self.set_primary_info()
        else:
            if possiblly_peakless:
                self.emg_peaks = []
                self.set_primary_info()
            else:
                self.get_emg_peaks(logger=self.logger)

        if USE_HALF_MAXIMUM_POINTS:
            if len(self.peak_info) == 1:
                half_spline = LSQUnivariateSpline( x, self.y_for_spline - self.max_y/2, self.knots[1:-1], ext=3 )
                self.half_x = half_spline.roots()
            else:
                self.half_x = None

        if add_features:
            try:
                self.feature = CurveFeatures(self)
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb)

        self.editor_ranges = None
        self.paired_ranges = None

        if debug_plot:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.KekLib.DebugPlot import get_parent
            fig = plt.figure( figsize=(16, 8) )
            proof_plot(self, get_parent(), fig)
            plt.show()

    def get_xy(self):
        # for forward compatibility
        return self.x, self.y

    def set_peak_dependent_info(self, cover_curve=None):
        # print('initial peak_top_x=', self.peak_top_x)
        work_boundaries = self.get_peak_boundaries()
        # print('work_boundaries=', work_boundaries)

        ranges, ptxes, valid_boundaries = self.get_ranges_by_ratio( PEAK_INFO_FIND_RATIO, boundaries=work_boundaries, return_ptxes=True)
        covered = False
        if len(ranges) == 0:
            if cover_curve is None:
                raise RuntimeError("No ranges")
            else:
                # as with Factin at wavelength=400
                # use the given cover curve to cover peak related info
                ranges = cover_curve.peak_info
                ptxes  = cover_curve.peak_top_x
                valid_boundaries = cover_curve.boundaries
                covered = True
                self.logger.warning("created an elution curve using a cover curve.")

        self.peak_top_x = ptxes
        # print('final peak_top_x=', self.peak_top_x)
        # print('valid_boundaries=', valid_boundaries)
        self.boundaries = valid_boundaries
        # self.boundaries = self.get_range_boudaries(ranges, valid_boundaries)

        # ------- backward compatibility attributes
        if not covered:
            # buggy for BSA_201
            # task: re-consider this
            self.modify_ranges_to_avoid_too_narrow_valley( ranges )
        # print('final ranges=', ranges)

        self.peak_info  = ranges
        if len(self.peak_info) == 0:
            from molass_legacy.KekLib.DebugPlot import get_parent
            fig = plt.figure( figsize=(16, 8) )
            proof_plot(self, get_parent(), fig)
            plt.show()
        assert len(self.peak_info) > 0
        self.set_major_infos()

    def get_emg_peaks(self, **kwargs):
        if self.emg_peaks is None:
            if False:
                from importlib import reload
                import DataStructure.EmgPeak
                reload(DataStructure.EmgPeak)
            from molass_legacy.DataStructure.EmgPeak import get_peaks
            self.emg_peaks = get_peaks(self, **kwargs)

        if len(self.emg_peaks) == 0 or self.possiblly_peakless:
            # case for UV elution at a longer wave length
            # len(self.emg_peaks) == 0 occurs for pH6
            self.set_primary_info()
            return self.emg_peaks

        # print('self.emg_peaks=', self.emg_peaks)
        # print('self.peak_info=', self.peak_info)

        # the following treatment is required for cases such as 20170309
        if len(self.emg_peaks) < len(self.peak_info):
            new_peak_top_x = []
            new_peak_info = []
            for k, peak in enumerate(self.peak_info):
                mid_x = peak[1]
                found = False
                for epeak in self.emg_peaks:
                    if abs(mid_x - epeak.top_x) < TOP_X_ALLOWANCE:
                        found = True
                        new_peak_top_x.append(self.peak_top_x[k])
                        new_peak_info.append(peak)
                        break
                if not found:
                    self.logger.info("peak %s without any corresponding emg_peak has been ignored.", str(peak))
            self.peak_top_x = np.array(new_peak_top_x)
            self.peak_info = new_peak_info
            self.boundaries = self.get_peak_boundaries()
        elif len(self.emg_peaks) > len(self.peak_info):
            new_emg_peaks = []
            for epeak in self.emg_peaks:
                top_x = epeak.top_x
                found = False
                for peak in self.peak_info:
                    if abs(top_x - peak[1]) < TOP_X_ALLOWANCE:
                        found = True
                        new_emg_peaks.append(epeak)
                        break
                if not found:
                    self.logger.info("emg_peak %s without any corresponding peak has been ignored.", str(epeak))
            self.emg_peaks = new_emg_peaks

        assert len(self.emg_peaks) == len(self.peak_info)

        self.set_primary_info()
        return self.emg_peaks

    def set_primary_info(self):
        primary_peak_no = None
        primary_peak    = None
        primary_peak_y  = None
        for i, info in enumerate(self.peak_info):
            px  = info[1]
            py  = self.spline( px )
            if primary_peak_y is None or py > primary_peak_y:
                primary_peak_no = i
                primary_peak    = px
                primary_peak_y  = py

        self.primary_peak_no    = primary_peak_no
        self.primary_peak_x     = primary_peak
        self.primary_peak_i     = None if primary_peak is None else int(primary_peak + 0.5)
        self.primary_peak_y     = primary_peak_y

        num_points = self.get_primary_peak_num_points()

        if False:

            if num_points is not None and num_points < BETER_SPLINE_LIM_PNTS:
                w = np.min(np.abs( self.knots - self.peak_top_x ))
                if w > BETER_SPLINE_LIM_WIDTH:
                    try:
                        top_x = self.peak_top_x[self.primary_peak_no]
                        lower_knots = self.knots[self.knots < top_x]
                        lower_knots[-1] = (lower_knots[-2] + top_x)/2
                        upper_knots = self.knots[self.knots > top_x]
                        upper_knots[0]  = (top_x + upper_knots[1])/2
                        self.knots = np.hstack( [lower_knots, [top_x], upper_knots] )
                        self.spline = LSQUnivariateSpline( x, self.y_for_spline, self.knots[1:-1], ext=3 )   # ext=3 is required for pH6
                        self.logger.info( "spline has been re-computed for this elution curve due to the small number(%d) of peak points" % num_points )
                    except:
                        self.logger.info("failed to get a better spline")

        # self.rich_peak_info = self.make_rich_peak_info(ranges)

    def get_primarypeak_x(self):
        return self.max_x

    def get_primarypeak_i(self):
        return int(round(self.max_x - self.x[0]))

    def get_primarypeak_i(self):
        return self.primary_peak_i

    def get_height_ratio(self, pos):
        return (self.spline(pos) - self.min_y) / self.height

    def get_major_valley_bottoms(self, major_peak_tops):
        bottoms = np.array(self.boundaries)
        ret_bottoms = []
        for k, top in enumerate(major_peak_tops):
            if k+1 < len(major_peak_tops):
                next_top = major_peak_tops[k+1]
                b = bottoms[np.logical_and(bottoms > top, bottoms < next_top)]
                # len(b) == 0 occured for 20181203 in FullOptInit
                if len(b) > 0:
                    ret_bottoms.append(b[0])

        return np.array(ret_bottoms)

    def set_major_infos(self):
        peak_rec_list = []
        peak_top_list = []
        for k, rec in enumerate(self.peak_info):
            if self.get_height_ratio(rec[1]) > MAJOR_PEAK_RATIO:
                peak_rec_list.append(rec)
                peak_top_list.append(self.peak_top_x[k])
        self.major_peak_info = peak_rec_list
        self.major_peak_top_x = np.array(peak_top_list)
        self.major_boundaries = self.get_major_valley_bottoms(self.major_peak_top_x)

    def get_major_peak_info(self, ratio=None, add_j0=False):
        if ratio is None:
            if add_j0:
                ret_info = [ [self.j0+i for i in rec] for rec in self.major_peak_info ]
            else:
                ret_info = self.major_peak_info
            return ret_info

        assert not add_j0

        peak_rec_list = []
        for k, rec in enumerate(self.peak_info):
            if self.get_height_ratio(rec[1]) > ratio:
                peak_rec_list.append(rec)

        return peak_rec_list

    def get_primary_peak_num_points(self):
        try:
            lower, center, upper = self.peak_info[self.primary_peak_no]
            num_points = upper - lower + 1
        except:
            num_points = None
        self.primary_peak_num_points = num_points
        return num_points

    def get_primary_peak_info(self):
        return self.peak_info[self.primary_peak_no]

    def has_few_points(self):
        return self.primary_peak_num_points < BETER_SPLINE_LIM_PNTS

    def determine_knots( self, i, x, low_quality=False ):
        if low_quality:
            num_nkots = NUM_KONTS_LOW_QUALITY
        else:
            if i == 0:
                num_nkots = min( MAX_NUM_KNOTS, max( MIN_NUM_KNOTS, len(x)//NUM_POINTS_PER_KNOT ) )
            else:
                num_nkots = NUM_KNOTS_MID_QUALITY
        return np.linspace( 0, len(x), num_nkots + 2 )

    def is_ignorable_peak(self, y):
        return y < self.max_y * PEAK_INFO_FIND_RATIO

    def get_significant_peaks( self, low_quality=False, unify_close_peaks=False ):
        self.roots = roots_ = self.smoothing.top_x
        return self.roots

    def low_peak_stands_out_clearly( self, root, peak_y, min_root_diff ):
        ratio = self.get_low_peak_stand_out_ratio( root, peak_y, min_root_diff )
        if ratio < LOW_PEAK_CHECK_LIMIT:
            return True

        ratio_w = self.get_low_peak_stand_out_ratio( root, peak_y, min_root_diff * 2 )
        improved_ratio = ratio_w/ratio
        # print( 'improved_ratio=', improved_ratio )

        return improved_ratio < 0.7

    def get_low_peak_stand_out_ratio( self, root, peak_y, min_root_diff ):
        peak_y_ = peak_y * LOW_PEAK_RAISE_RATIO     # to allow low peaks as in Sugiyama

        start   = int( max(0, root - min_root_diff ) + 0.5 )
        stop    = int( min(len(self.x), root + min_root_diff ) + 0.5 )
        slice_  = slice( start, stop )
        y_  = self.y[slice_]
        ratio   = len( np.where( y_ > peak_y_ )[0] ) / (stop - start)
        if False:
            print( 'exceed ratio=', ratio )
            x_  = self.x[slice_]
            plt.plot( x_, y_ )
            plt.plot( x_, self.spline( x_ ) )
            ppp = np.where( y_ > peak_y_ )[0]
            plt.plot( x_[ppp], y_[ppp], 'o', color='yellow' )
            plt.plot( root, peak_y, 'o', color='red' )
            plt.show()
        return ratio

    def get_peak_boundaries(self, peak_top_x=None, debug=False):

        if peak_top_x is None:
            peak_top_x = self.peak_top_x

        last_ptx = None
        ret_boundaries = []
        if len(peak_top_x) == 1:
            # temporary fix for PgpTak in V2
            return ret_boundaries

        print('get_peak_boundaries: peak_top_x=', peak_top_x)
        for ptx in peak_top_x:
            if last_ptx is not None:
                start   = int(last_ptx)
                stop    = int(ptx)
                slice_ = slice( start, stop )
                x_  = self.x[slice_]
                y_  = self.spline(x_)
                n   = np.argmin( y_ )
                try:
                    kth = min(6, int( len(y_) * 0.1 ))
                    y_kth = np.argpartition( y_, kth )
                    n_  = int( np.average( y_kth[0:kth] ) + 0.5 )
                    y_lim = y_[n]*1.3
                    if y_lim > 0 and y_[n_] < y_lim:
                        # to avoid cases like 20161006_quality
                        n   = n_
                    print('try success: kth=', kth, 'y_kth[0:kth]=', y_kth[0:kth], 'n_=', n_)
                except:
                    # nothing to do?
                    pass

                ret_boundaries.append( start + n )

            last_ptx = ptx

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            x = self.x
            y = self.y
            with plt.Dp():
                fig = plt.figure()
                ax = fig.gca()
                ax.set_title("get_peak_boundaries debug")
                ax.plot(x, y, label='data')
                ax.plot(x, self.spline(x), label='spline')
                ax.plot(peak_top_x, self.spline(peak_top_x), 'o', color='red', label='peak top')
                ax.plot(ret_boundaries, self.spline(ret_boundaries), 'o', color='green', label='boundaries')
                ax.legend()
                fig.tight_layout()
                plt.show()

        if len(peak_top_x) > 0 and len(ret_boundaries) != len(peak_top_x) - 1:
            print('ret_boundaries=', ret_boundaries)
            print('peak_top_x=', peak_top_x)
            assert len(ret_boundaries) == len(peak_top_x) - 1

        return ret_boundaries

    def get_lower_from_y(self, start, pti, yr, debug=False):
        try:
            # y_ = self.sy[start:pti] if self.low_quality else self.y[start:pti]
            y_ = self.sy[start:pti]
            lower0  = start + np.where( y_ < yr )[0][-1]
        except:
            # in case there exists no such points
            lower0  = start

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            print("pti=", pti)
            x = self.x
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_lower_from_y")
                ax.plot(x, y)
                ax.plot(x, self.sy, ":")
                xmin, xmax = ax.get_xlim()
                ax.set_xlim(xmin, xmax)
                ax.plot([xmin, xmax], [yr, yr], color="yellow")
                for j in [start, pti]:
                    px = x[j]
                    ax.plot(px, self.spline(px), "o")
                x_ = x[lower0]
                ax.plot(x_, self.spline(x_), "o", color="red")
                fig.tight_layout()
                plt.show()

        return lower0

    def get_lower_from_d1y( self, start, pti ):
        try:
            lower1  = start + np.where( self.d1y[start:pti] < 0 )[0][-1]
        except:
            # in case there exists no such points
            lower1  = start
        return lower1

    def get_upper_from_y(self, stop, pti, yr, debug=False):
        try:
            # y_ = self.sy[pti:stop] if self.low_quality else self.y[pti:stop]
            y_ = self.sy[pti:stop]
            upper0  = pti + np.where( y_ < yr )[0][0]
        except:
            # in case there exists no such points
            upper0  = stop - 1
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            x = self.x
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_upper_from_y")
                ax.plot(x, y)
                ax.plot(x, self.sy, ":")
                xmin, xmax = ax.get_xlim()
                ax.set_xlim(xmin, xmax)
                ax.plot([xmin, xmax], [yr, yr], color="yellow")
                for p in [stop-1, pti]:
                    px = x[p]
                    ax.plot(px, self.spline(px), "o")
                x_ = x[upper0]
                ax.plot(x_, self.spline(x_), "o", color="red")
                fig.tight_layout()
                plt.show()
        return upper0

    def get_upper_from_d1y( self, stop, pti ):
        try:
            # let slice start at pti+1 to avoid taking pti as upper1
            upper1  = pti + np.where( self.d1y[pti+1:stop] > 0 )[0][0]
        except:
            # in case there exists no such points
            upper1  = stop - 1
        return upper1

    def get_default_editor_ranges(self):
        if self.editor_ranges is None:
            ranges = []
            for prec, epeak in zip(self.peak_info, self.emg_peaks):
                top = prec[1]
                left, right = epeak.get_model_x_from_ratio(ANALYSIS_RANGE_RATIO)
                left_ = max(0, int(round(left)))
                right_ = min(len(self.x)-1, int(round(right)))
                if left_ < top and top < right_ and right_ - left_> MIN_HALF_WIDTH*2:
                    range_ = [[left_, top], [top, right_]]
                else:
                    left_c = top - MIN_HALF_WIDTH
                    right_c = top + MIN_HALF_WIDTH
                    self.logger.warning("illegal (or too narrow) range values (%d, %d, %d) corrected to (%d, %d) as a temporary fix.", left_, top, right_, left_c, right_c)
                    range_  = [[left_c, right_c]]
                ranges.append(range_)
            self.editor_ranges = ranges
        return self.editor_ranges

    def get_default_paired_ranges(self):
        if self.paired_ranges is None:
            paired_ranges = []
            for k, erec in enumerate(self.get_default_editor_ranges()):
                left, top = erec[0]
                pinfo = PeakInfo(k, top)
                paired_ranges.append(PairedRange(pinfo, *erec))     # erec = [ [left, top], [top, right] ]
            self.paired_ranges = paired_ranges
        return self.paired_ranges

    def get_ranges_by_ratio( self, ratio, boundaries=None, return_ptxes = False, debug=False ):
        if boundaries is None:
            boundaries = self.boundaries
            regular = True
        else:
            regular = False

        # boundaries_ should be integers. why are non-integers coming in?
        boundaries_ = np.array(boundaries)

        # debug = get_setting("local_debug")

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            x = self.x
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_ranges_by_ratio begin")
                ax.plot(x, y)
                for b in boundaries_:
                    ax.plot(b, self.spline(b), "o", color="green")
                fig.tight_layout()
                plt.show()

        ranges = []
        ptxes = []
        valid_boundaries = []
        start = 0

        for k, ptx in enumerate(self.peak_top_x):
            pti = int(ptx - self.x[0] + 0.5)

            if k < len(boundaries_):
                if regular:
                    stop = int(boundaries_[k])
                else:
                    try:
                        stop = int(boundaries_[boundaries_ > ptx][0])
                    except:
                        stop = len(self.y)
            else:
                stop = len(self.y)

            yr  = self.spline(ptx) * ratio
            lower   = self.get_lower_from_y( start, pti, yr, debug=debug )
            upper   = self.get_upper_from_y( stop, pti, yr, debug=debug )
            width   = upper - lower

            if debug:
                print("peak_top_x=", self.peak_top_x)
                print("boundaries_=", boundaries_)
                print("[lower, pti, upper]=",  [lower, pti, upper] )
                x = self.x
                y = self.y
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("[%d] range" % k)
                    ax.plot(x, y)
                    for p in [lower, upper]:
                        px = x[p]
                        ax.plot(px, self.spline(px), "o")
                    fig.tight_layout()
                    plt.show()

            if debug and (lower >= pti or pti >= upper):
                print('regular=', regular)
                print('boundaries=', boundaries)
                print('start=', start, 'stop=', stop)
                print( 'width=', width, 'min_peak_width=', self.min_peak_width )
                fig = plt.figure()
                ax = fig.gca()
                ax.cla()
                ax.set_title('[%d] ptx=%g' % (k, ptx))
                ax.plot( self.x, self.y )
                top_x = self.peak_top_x
                ax.plot(top_x, self.spline(top_x), 'o', color='orange' )

                b = boundaries
                if b is not None and len(b) > 0:
                    ax.plot(self.x[b], self.y[b], 'o', color='green' )

                colors = ['yellow', 'red', 'cyan']
                for j, x_ in enumerate([lower, ptx, upper]):
                    c = colors[j]
                    ax.plot( x_, self.spline(x_), 'o', color=c )
                plt.show()

            height_ratio = self.spline(ptx)/self.max_y
            def major_height_ratio_ok():
                top_x = (lower + upper)//2
                dw = width//3
                ratio = np.average(self.y[top_x-dw:top_x+dw])/self.max_y
                self.logger.info("adopted a peak at %d with average_ratio=%.3g", top_x, ratio)
                return ratio > 0.1

            if height_ratio > SUFFICIENT_PEAK_H_RATIO or width >= self.min_peak_width or major_height_ratio_ok():
                if (    lower >= 0 and lower <= pti
                    and pti <= upper and upper < len(self.x)
                    ):
                    if len(ranges) > 0:
                        last_upper = ranges[-1][2]
                        if lower - last_upper < 2:
                            # as in 20191118_3
                            last_range = copy.deepcopy(ranges[-1])
                            ranges[-1][2] = last_upper - 1
                            self.logger.warning("ranges modified as a temporary fix from %s to %s, %s to %s",
                                                str(last_range), str(ranges[-1]), str([lower, pti, upper]), str([lower+1, pti, upper]))
                            lower += 1
                    ranges.append( [lower, pti, upper] )
                    ptxes.append( ptx )
                    if len(ranges) > 1:
                        valid_boundaries.append(start)  # i.e., start=last stop
                else:
                    self.logger.warning("ignored irregular range " + str([lower, pti, upper]) + " in get_ranges_by_ratio")
            else:
                if width < 0:
                    self.logger.warning("negative peak width %d = %d - %d" % (width, upper, lower))
                else:
                    self.logger.warning("too small peak width %d = %d - %d" % (width, upper, lower))

                # if len(valid_boundaries) > 0:
                #    valid_boundaries.pop()

            start   = stop

        if debug:
            print("---- ret ranges=", ranges)
            print("---- ret ptxes=", ptxes)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_ranges_by_ratio")
                ax.plot(self.x, self.y)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for range_ in ranges:
                    for k, j in enumerate(range_):
                        x_ = self.x[j]
                        ax.plot([x_, x_], [ymin, ymax], ":", color="yellow")
                        ax.plot(x_, self.spline(x_), "o", color="C%d" % (k+1))
                fig.tight_layout()
                plt.show()

        if len(ranges) == 0:
            assert len(valid_boundaries) == 0
            ranges, ptxes = self.make_just_one_peak_info(ratio)

        assert len(ranges) == len(ptxes)

        if len(ranges) != len(valid_boundaries) + 1:
            print('ranges=', ranges, 'valid_boundaries=', valid_boundaries)
            fig = plt.figure()
            ax = fig.gca()
            simple_plot(ax, self, "get_ranges_by_ratio end", boundaries=valid_boundaries, spline=True)
            plt.show()

        assert len(ranges) == len(valid_boundaries) + 1

        if return_ptxes:
            return ranges, np.array(ptxes), valid_boundaries
        else:
            return ranges

    def make_just_one_peak_info(self, ratio):
        # a_curve2 for 20190221_1_ver2

        n = np.argmax(self.spline(self.peak_top_x))
        ptx = self.peak_top_x[n]
        pti = int(ptx - self.x[0] + 0.5)
        yr  = self.spline(ptx) * ratio
        lower   = self.get_lower_from_y( 0, pti, yr )
        upper   = self.get_upper_from_y( len(self.x), pti, yr )

        peak_rec = [lower, pti, upper]
        self.logger.warning("made just one peak %g, %s to avoid creating a no-peak curve instance." % (ptx, str(peak_rec)))
        return [peak_rec], [ptx]

    def modify_ranges_to_avoid_too_narrow_valley(self, ranges, debug=False):
        min_width = min(MININUM_PEAK_WIDTH, int(len(self.x)*MININUM_PEAK_W_RATIO/2))

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            print("min_width=", min_width)
            print("ranges before=", ranges )
            print("self.boundaries=", self.boundaries )
            ranges_save = copy.deepcopy(ranges)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("modify_ranges_to_avoid_too_narrow_valley debug")
                ax.plot(self.x, self.y)
                for rec in ranges:
                    for k, j in enumerate(rec):
                        ax.plot(j, self.spline(j), "o", color="C%d" % (k+1))
                ax.plot(self.boundaries, self.spline(self.boundaries), "o", color="green")
                fig.tight_layout()
                plt.show()

        if len(ranges) == len(self.boundaries) + 1:
            boundaries = self.boundaries
        else:
            boundaries = self.get_range_boudaries(ranges, self.boundaries)

        num_changes = 0
        for k, b in enumerate(boundaries):
            n = b - ranges[k][2]
            if n < min_width:
                ranges[k][2] -= ( min_width - n )
                num_changes += 1
            n = ranges[k+1][0] - b
            if n < min_width:
                ranges[k+1][0] += ( min_width - n )
                num_changes += 1

        if num_changes > 0 and debug:
            self.logger.warning('%d in points %s were changed to avoid too narrow valleys resulting in %s.', num_changes, str(ranges_save), str(ranges))
            # too many to log

    def get_range_boudaries(self, ranges, boundaries):
        """
        TODO: investigate this case in pH7
        """
        print( 'ranges=', ranges )
        print( 'boundaries=', boundaries )

        ret_boundaries = []
        start = 0
        for k, range_ in enumerate(ranges):
            lower1, top1, upper1 = range_
            if k < len(ranges) - 1:
                next_range = ranges[k+1]
                lower2, top2, upper2 = next_range
                left_lim = (top1 + upper1)/2
                right_lim = (lower2 + top2)/2
                for b in boundaries[start:]:
                    if left_lim < b and b < right_lim:
                        if upper1 > b:
                            range_[2] = b
                        if b > lower2:
                            next_range[0] = b
                        ret_boundaries.append(b)
                        start = k
                        break

        if len(ranges) > 0 and len(ret_boundaries) != len(ranges) - 1:
            print(ranges)
            print(boundaries)
            print(ret_boundaries)
            assert len(ret_boundaries) != len(ranges) - 1

        print( 'ret_boundaries=', ret_boundaries )

        return ret_boundaries

    def remove_peaks( self, to_remove ):
        if len(to_remove) == 0:
            return

        # required for 20190302-analysis_copy
        new_top_x = []
        for k, x in enumerate(self.peak_top_x):
            if k not in to_remove:
                new_top_x.append(x)

        self.peak_top_x = np.array(new_top_x)
        self.set_peak_dependent_info()
        self.logger.warning("peaks %s have been removed.", str(to_remove))

    def add_peaks( self, to_add ):
        print( 'to_add=', to_add )
        if len(to_add) == 0:
            return
        # TODO
        assert False

    def make_rich_peak_info(self, ranges):
        """
        experimental
        """
        rich_peak_info = copy.deepcopy(ranges)

        x = self.x
        y = self.smoothed_y

        foot_level = FOOT_LEVEL_RATIO * self.max_y
        shifted_spline = LSQUnivariateSpline(x, y-foot_level, self.knots[1:-1], ext=3)
        candidate_foots = shifted_spline.roots()
        # print('candidate_foots=', candidate_foots)

        start_peak = 0
        foots = [[]]*len(ranges)
        for f in candidate_foots:
            for k, info in enumerate(ranges[start_peak:], start=start_peak):
                left, _, right = info
                if left - FOOT_ADD_WIDTH < f and f < right + FOOT_ADD_WIDTH:
                    foots[k].append(f)
                    start_peak = k
                    break

        print('foots=', foots)
        foots_ = np.array(foots).flatten()

        if False:
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(x, y, label='data')
            ax.plot(x, self.spline(x), label='spline')
            # ax.plot(x, shifted_spline(x), label='shifted spline')
            ax.plot(candidate_foots, self.spline(candidate_foots), 'o', color='pink', label='candidate_foots')
            ax.plot(foots_, self.spline(foots_), 'o', color='red', label='foots')
            ax.legend()
            plt.show()

        return rich_peak_info

    def locally_tall_enough(self, pos):
        delta_x = self.smoothing.delta_x
        delta_y = self.smoothing.delta_y

        py = self.spline(pos)
        left_h  = py - self.spline(pos-delta_x)
        right_h = py - self.spline(pos+delta_x)
        local_height = (left_h + right_h)/2
        return local_height > delta_y

    def locally_tall_enough_evenif_rotated(self, pos, delta_x=None):
        # TODO: check after rotation

        if delta_x is None:
            delta_x = self.smoothing.delta_x
        delta_y = self.smoothing.delta_y

        py = self.spline(pos)
        left_h  = py - self.spline(pos-delta_x)
        right_h = py - self.spline(pos+delta_x)
        local_height = (left_h + right_h)/2
        return local_height > delta_y

    def compute_valley_bottom_score(self, k):
        b = self.boundaries[k]
        bottom_y = max(0, self.spline(b))
        top_y_left  = self.peak_info[k][1]
        top_y_right = self.peak_info[k+1][1]

        left_ratio = bottom_y/top_y_left
        right_ratio = bottom_y/top_y_right

        ratio = min(1, (left_ratio + right_ratio)/2 )
        return 1 - ratio

    def compute_moments(self):
        if self.mean is None:
            x = self.x
            y = np.zeros(len(x))

            # self.y can be negative as in 20200123_1
            positive_y = self.y > 0
            y[positive_y] = self.y[positive_y]

            sum_y = np.sum(y)
            mean  = np.sum(x*y)/sum_y
            variance = np.sum(x**2*y)/sum_y - mean**2
            self.mean = mean
            self.variance = variance
        return self.mean, self.variance

    def get_end_y(self, slice_):
        y_ = self.y[slice_]
        h = abs(y_[-1] - y_[0])
        ok = False
        if h/self.max_y < END_DIFF_MIN_HEIGHT:
            ok = True
        else:
            std = np.std(y_)
            if std/self.max_y < END_DIFF_MIN_HEIGHT:
                ok = True
        if ok:
            self.end_slices.append(slice_)
            return np.average(y_)

        # in this case, slice_ must be narrowed as obsesrved in 20181204
        if slice_.start == 0:
            return self.get_end_y(slice(0, slice_.stop//2))
        else:
            return self.get_end_y(slice(slice_.start//2, None))

    def compute_sedimentation_rate(self, debug=False):
        emg_peaks = self.get_emg_peaks()
        self.end_slices = []
        start_y = self.get_end_y(slice(0, END_DIFF_WIDTH))
        end_y = self.get_end_y(slice(-END_DIFF_WIDTH, None))
        sedrate = (end_y - start_y)/(self.max_y - start_y)

        slice1 = self.end_slices[0]
        width1 = slice1.stop

        slice2 = self.end_slices[1]
        width2 = abs(slice2.start)
        start2 = len(self.x) + slice2.start

        num_emg_peaks = len(emg_peaks)

        if num_emg_peaks > 0:
            # to reduce the calulation
            indeces = set([0, num_emg_peaks-1])

            lim_x_pairs = []
            for i in indeces:
                if USE_RELIABLE_ALPHA:
                    pair = emg_peaks[i].get_model_x_from_ratio(RELIABLE_ALPHA)
                else:
                    pair = emg_peaks[i].get_sigma_points(RELIABLE_SIGMA_RATIO)
                lim_x_pairs.append(pair)
            sigma_x1 = lim_x_pairs[0][0]
            sigma_x2 = lim_x_pairs[-1][1]
            reliable = (width1 >= END_DIFF_WIDTH//8 and slice1.stop < sigma_x1) and (width2 >= END_DIFF_WIDTH//8 and sigma_x2 < start2)
            if debug:
                import molass_legacy.KekLib.DebugPlot as plt
                print(self.end_slices, sigma_x1, sigma_x2, start2)
                print(width1, width2)

                x = self.x
                y = self.y

                plt.push()
                fig, ax = plt.subplots()
                ax.set_title("compute_sedimentation_rate debug")
                ax.plot(x, y, label='data')
                for slice_ in [slice(0,END_DIFF_WIDTH), slice(-END_DIFF_WIDTH, None)]:
                    ax.plot(x[slice_], y[slice_], 'o', color='yellow')

                for peak in emg_peaks:
                    print(self.j0, peak, peak.get_sigma_points(RELIABLE_SIGMA_RATIO))

                for k in set([0, len(emg_peaks)-1]):
                    peak = emg_peaks[k]
                    ax.plot(x, peak.get_model_y(x), label='model')

                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for sx in [sigma_x1, sigma_x2]:
                    ax.plot([sx, sx], [ymin, ymax], ':', color='gray')

                ax.legend()
                fig.tight_layout()
                plt.show()
                plt.pop()
        else:
            # temporary fix for pH6 to avoid a bug with the revised EmgPeak
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "compute_sedimentation_rate failure: ", n=10)
            reliable = False

        self.logger.info("computed end_slices as %s, sedimentation rate as %.3g which is%s reliable", str(self.end_slices), sedrate, "" if reliable else " not")
        return sedrate, reliable

    def get_end_slices(self):
        if self.end_slices is None:
            # this case is not desirable.
            # improve so that self.compute_sedimentation_rate be called only once.
            self.compute_sedimentation_rate()
        return self.end_slices

    def peak_no_from_range(self, start, stop):
        max_width = None
        ret_no = None
        for k, rec in enumerate(self.peak_info):
            lsp, _, rsp = rec
            width = max(0, min(rsp, stop) - max(lsp, start))
            if width > 0:
                if max_width is None or width > max_width:
                    ret_no = k
                    max_width = width
        return ret_no

    def get_end_points(self, debug=False):
        from molass_legacy.KekLib.GeometryUtils import rotate
        x = self.x
        y = self.y
        pinfo1 = self.peak_info[0]
        position_ratio = pinfo1[1]/len(x)
        if position_ratio < 0.05:
            pinfo1 = self.peak_info[1]
        pinfo2 = self.peak_info[-1]
        ret_points = []
        for k, info in enumerate([pinfo1, pinfo2]):
            m = info[1]
            if k == 0:
                slice_ = slice(0, m)
                angle = -np.pi/4        # 
                ty = y[m]
            else:
                slice_ = slice(m, None)
                angle = np.pi/4
                ty = y[m]
            x_ = self.x[slice_]
            wx = (x_ - x_[0]) / len(x_)
            h_ = y[m]
            y_ = y[slice_]
            wy = y_ / h_
            rx, ry = rotate(angle, wx, wy)
            n = np.argmin(ry)
            if n > 0 and n < len(ry) - 1:
                n_ = int(x_[0]) + n
                px = x[m]
                py = y[m]
                fx = x[n_]
                fy = y[n_]
                gx = px + (fx - px)*ty/(py - fy)
            else:
                if k == 0:
                    n = 0
                    fx = x_[n]
                    fy = y_[n]
                    gx = fx
                else:
                    # as in 20161119/Kosugi8
                    n = -1
                    fx = x_[n]
                    fy = y_[n]
                    gx = fx
            ret_points.append(gx)

            if debug:
                import molass_legacy.KekLib.DebugPlot as plt
                plt.push()
                fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(21,6))
                ax0.plot(x_, y_)
                ax0.plot(fx, fy, 'o', color='red')
                ax0.plot(gx, py - ty, 'o', color='yellow')
                ax1.plot(wx, wy)
                ax1.plot(wx[n], wy[n], 'o', color='red')
                ax2.plot(rx, ry)
                ax2.plot(rx[n], ry[n], 'o', color='red')
                fig.tight_layout()
                plt.show()
                plt.pop()

        return ret_points

    def get_peak_position(self, start, stop):
        j = (start + stop)/2
        min_pos = None
        min_val = None
        for rec in self.peak_info:
            pos = rec[1]
            val = abs(pos - j)
            if min_val is None or val < min_val:
                min_val = val
                min_pos = pos
        self.logger.info("pos=%d for range(%d, %d)", min_pos, start, stop)
        return min_pos

    def get_model_param_list(self):
        return [epeak.get_params() for epeak in self.get_emg_peaks()]

    def get_peak_region_sigma(self):
        if self.peak_region_sigma is None:
            emg_peaks = self.get_emg_peaks()
            h1, m1, s1, t1 = emg_peaks[0].get_params()
            h2, m2, s2, t2 = emg_peaks[-1].get_params()
            peak_region = (max(0, m1-3*s1), min((len(self.x)), m2+3*s2))
            self.peak_region_sigma = peak_region
        return self.peak_region_sigma

    def get_peak_region(self, sigma_scale=3):
        emg_peaks = self.get_emg_peaks()
        h1, m1, s1, t1 = emg_peaks[0].get_params()
        h2, m2, s2, t2 = emg_peaks[-1].get_params()
        peak_region = (max(0, m1-sigma_scale*s1), min((len(self.x)), m2+sigma_scale*s2))
        return peak_region

    def get_peak_region_width(self):
        peak_region = self.get_peak_region_sigma()
        return peak_region[1] - peak_region[0]
    
    def is_roughly_centered(self):
        x = self.x
        xc = (x[0] + x[-1])/2
        return abs(xc - self.max_x)/(x[-1] - x[0]) < 0.3


"""
    _MOLASS.DummyClasses.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
class Dummy:
    def __repr__(self):
        return "%s()" % self.__class__.__qualname__

class DummySd(Dummy):
    def __init__(self):
        self.mc_vector = None

class DummyMapper(Dummy):
    def __init__(self):
        self.x_ranges = None
 
class DummyJudgeHolder(Dummy):
    def __init__(self):
        pass

class PeakInfo(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class DecompEditorInfo(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class RangeEditorInfo(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class PairedRange(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class PreviewData(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class PreviewOptions(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class MappingParams(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class TrimmingInfo(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class ModelEvaluator(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class DualEvaluator(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class EGHA(Dummy):
    def __init__(self, *args, **kwargs):
        pass

class EMGA(Dummy):
    def __init__(self, *args, **kwargs):
        pass

if __name__ == '__main__':
    print(DummySd())



# coding: utf-8
"""
    MethodAdjusterFrame.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk

class MethodAdjusterFrame(Tk.Frame):
    def __init__(self, parent):
        Tk.Frame.__init__(self, parent)
        self.body(self)

    def body(self, body_frame):
        space = Tk.Frame(body_frame, height=300)
        space.pack()
        label = Tk.Label(body_frame, text="In Preparation", font=('', 40), fg='gray')
        label.pack()

    def body_trash(self, body_frame):
        from PIL import Image, ImageTk
        dir_, _ = os.path.split(__file__)
        path = os.path.join(dir_, 'MethodFileAdjuster.png')
        img = Image.open(path)
        img = ImageTk.PhotoImage(img)
        canvas = Tk.Canvas(body_frame, bg='black', width=400, height=300)
        canvas.pack()
        canvas.create_image(0, 0, image=img, anchor=Tk.NW)



# coding: utf-8
"""
    QmmOptions.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import re
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

NUM_SEEDS = 4

class QmmOptionsDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, title="QMM Options", visible=False, location='lower right')

    def show(self):
        self._show()

    def body(self, bframe):
        iframe = Tk.Frame(bframe)
        iframe.pack(padx=20, pady=10)

        grid_row = 0
        label = Tk.Label(iframe, text="Number of components")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        self.auto_n_components = Tk.IntVar()
        self.auto_n_components.set(get_setting('auto_n_components'))
        auto_cb = Tk.Checkbutton(iframe, text="automatic", variable=self.auto_n_components)
        auto_cb.grid(row=grid_row, column=1, sticky=Tk.W, padx=5 )

        grid_row += 1
        self.n_components = Tk.IntVar()
        self.n_components.set(get_setting('n_components'))

        self.n_components_spinbox = Tk.Spinbox( iframe, textvariable=self.n_components,
                                from_=1, to=16, increment=1,
                                justify=Tk.CENTER, width=6)

        self.n_components_spinbox.grid(row=grid_row, column=1, sticky=Tk.W, padx=5)

        self.auto_n_components_tracer()
        self.auto_n_components.trace("w", self.auto_n_components_tracer)

        grid_row += 1
        space = Tk.Frame(iframe, height=5)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Number of iterations")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        self.max_iterations = Tk.IntVar()
        self.max_iterations.set(get_setting('max_iterations'))
        spinbox = Tk.Spinbox( iframe, textvariable=self.max_iterations,
                                from_=10, to=1000, increment=100,
                                justify=Tk.CENTER, width=6)

        spinbox.grid(row=grid_row, column=1, sticky=Tk.W, padx=5, pady=5)

        self.conc_dependence = Tk.IntVar()
        cd = get_setting('conc_dependence')
        self.conc_dependence.set(cd)

        grid_row += 1
        label = Tk.Label(iframe, text="Conc. dependence")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        spinbox = Tk.Spinbox( iframe, textvariable=self.conc_dependence,
                                from_=1, to=2, increment=1,
                                justify=Tk.CENTER, width=6)

        spinbox.grid(row=grid_row, column=1, sticky=Tk.W, padx=5)

        grid_row += 1
        label = Tk.Label(iframe, text="QMM fitting mode")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        self.qmm_separately = Tk.IntVar()
        self.qmm_separately.set(get_setting('qmm_separately'))
        mode_cb = Tk.Checkbutton(iframe, text="separately", variable=self.qmm_separately)
        mode_cb.grid(row=grid_row, column=1, sticky=Tk.W, padx=5)

        grid_row += 1
        label = Tk.Label(iframe, text="Random seeds")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        self.random_seed_fix = Tk.IntVar()
        fixed_random_seeds = get_setting('fixed_random_seeds')
        self.random_seed_fix.set(0 if fixed_random_seeds is None else 1)
        seeds_cb = Tk.Checkbutton(iframe, text="fixed", variable=self.random_seed_fix)
        seeds_cb.grid(row=grid_row, column=1, sticky=Tk.W, padx=5 )

        grid_row += 1
        random_seed_frame = Tk.Frame(iframe)
        random_seed_frame.grid(row=grid_row, column=1, sticky=Tk.W, padx=5)
        self.random_seeds = Tk.StringVar()
        if fixed_random_seeds is None:
            fixed_random_seeds = get_setting('last_random_seeds')
            if fixed_random_seeds is None:
                import numpy as np
                fixed_random_seeds = np.random.randint(1000, 9999, NUM_SEEDS)

        self.random_seeds.set( ' '.join(['%4d' % seed for seed in fixed_random_seeds]) )
        self.seed_entry = Tk.Entry(random_seed_frame, textvariable=self.random_seeds, width=19, justify=Tk.CENTER)
        self.seed_entry.pack()

        self.random_seed_fix.trace("w", self.random_seed_fix_tracer)
        self.random_seed_fix_tracer()

        grid_row += 1
        label = Tk.Label(iframe, text="Denoise rank")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        forced_denoise_rank = get_setting('forced_denoise_rank')
        self.auto_denoise_rank = Tk.IntVar()
        self.auto_denoise_rank.set(1 if forced_denoise_rank is None else 0)
        rank_cb = Tk.Checkbutton(iframe, text="automatic", variable=self.auto_denoise_rank)
        rank_cb.grid(row=grid_row, column=1, sticky=Tk.W, padx=5 )

        grid_row += 1
        if forced_denoise_rank is None:
            denoise_rank = get_setting('last_denoise_rank')
        else:
            denoise_rank = forced_denoise_rank
        self.denoise_rank = Tk.IntVar()
        self.denoise_rank.set(denoise_rank)

        spinbox = Tk.Spinbox( iframe, textvariable=self.denoise_rank,
                                from_=1, to=32, increment=1,
                                justify=Tk.CENTER, width=6)

        spinbox.grid(row=grid_row, column=1, sticky=Tk.W, padx=5)
        self.denoise_rank_sbox = spinbox

        self.auto_denoise_rank.trace("w", self.auto_denoise_rank_tracer)
        self.auto_denoise_rank_tracer()

        grid_row += 1
        label = Tk.Label(iframe, text="Rg estimation")
        label.grid(row=grid_row, column=0, sticky=Tk.W, pady=5)

        denss_fitted_rg = get_setting('denss_fitted_rg')
        self.denss_fitted_rg = Tk.IntVar()
        self.denss_fitted_rg.set(denss_fitted_rg)
        fitted_rg_cb = Tk.Checkbutton(iframe, text="apply fit_data (DENSS)", variable=self.denss_fitted_rg)
        fitted_rg_cb.grid(row=grid_row, column=1, sticky=Tk.W, padx=5 )

    def auto_n_components_tracer(self, *args):
        auto = self.auto_n_components.get()
        self.n_components_spinbox.config(state=Tk.DISABLED if auto else Tk.NORMAL)

    def random_seed_fix_tracer(self, *args):
        fix = self.random_seed_fix.get()
        state = Tk.NORMAL if fix else Tk.DISABLED
        self.seed_entry.config(state=state)

    def auto_denoise_rank_tracer(self, *args):
        auto = self.auto_denoise_rank.get()
        self.denoise_rank_sbox.config(state=Tk.DISABLED if auto else Tk.NORMAL)

    def validate(self):
        try:
            self.input_random_seeds = [int(seed) for seed in re.split(r'\s+', self.random_seeds.get())]
            self.seed_entry.config(fg='black')
            ret = True
        except:
            self.seed_entry.config(fg='red')
            ret = False
        return ret

    def apply(self):
        set_setting('auto_n_components', self.auto_n_components.get())
        set_setting('n_components', self.n_components.get())
        set_setting('max_iterations', self.max_iterations.get())
        set_setting('conc_dependence', self.conc_dependence.get())
        set_setting('qmm_separately', self.qmm_separately.get())
        if self.random_seed_fix.get():
            fixed_random_seeds = self.input_random_seeds
        else:
            fixed_random_seeds = None
        set_setting('fixed_random_seeds', fixed_random_seeds)

        auto_denoise_rank = self.auto_denoise_rank.get()
        if auto_denoise_rank:
            denoise_rank = None
        else:
            denoise_rank = self.denoise_rank.get()
        set_setting('forced_denoise_rank', denoise_rank)
        set_setting('denss_fitted_rg', self.denss_fitted_rg.get())



# coding: utf-8
"""
    Baseline.LpmInspect.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import copy
import logging
import numpy as np
from bisect import bisect_right
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import SpanSelector
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.DENSS.DenssUtils import fit_data
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from MatrixData import simple_plot_3d
from .Baseline import compute_baseline, better_integrative_curve, END_WIDTH
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from DataUtils import cut_upper_folders
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
import tkinter.ttk as ttk
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from .BaselineGuinier import BaselineGuinier, get_peaktop_slice

TITLE_FONTSIZE = 18
LEGEND_FONTSIZE = 14
USE_OPTINAL_BUTTONS = False
DEBUG = False

class EcurveProxyCds:
    def __init__(self, ecurve, j_slice):
        self.ecurve = ecurve
        self.j_start = j_slice.start
        if self.j_start is None:
            self.j_start = 0
        self.y = ecurve.y[j_slice]

    def get_major_peak_info(self):
        ret_info = []
        for rec in self.ecurve.get_major_peak_info():
            ret_info.append([j - self.j_start for j in rec])
        return ret_info

def demo(root, in_folder, md, integral=False, for_all_q=True, peakno=None, smooth_line=True, show_preview=False):
    print(in_folder)

    if show_preview:
        from molass_legacy.Tools.ThreeDimViewer import ThreeDimViewer
        dialog = ThreeDimViewer(root, md)
        dialog.show()

    fig = dplt.figure(figsize=(21, 11))
    demo_impl(fig, in_folder, md, integral=integral, for_all_q=for_all_q, peakno=peakno, smooth_line=smooth_line, show_lpm_fixed=True)
    dplt.show()

def demo_impl(fig, in_folder, md, j0=0, integral=False, for_all_q=True, peakno=None,
                smooth_line=False, show_scd=False,
                show_lpm_fixed=False, suppress_titles=False, end_slices=None, logger=None):

    xr = md.xr

    if False:
        xr.set_elution_curve()

    i_slice = slice(0, xr.i_slice.stop)
    j_slice = xr.j_slice
    i_index = xr.e_index - i_slice.start

    whole_data = xr.data.copy()
    data = whole_data[i_slice,j_slice]
    error = xr.error.data[i_slice,j_slice]
    q = xr.vector[i_slice]

    if peakno is None:
        peakno = xr.e_curve.primary_peak_no

    if show_scd:
        from Conc.ConcDepend import ConcDepend
        ecurve_for_scd = EcurveProxyCds(xr.e_curve, j_slice)
        cd = ConcDepend(q, data, error, ecurve_for_scd)
        scd_list = cd.compute_judge_info()
        scd0 = scd_list[peakno][1]

    # z = xr.e_curve.y[j_slice]
    z = data[i_index,:]

    if False:
        x = xr.e_curve.x
        y = xr.e_curve.y
        j0 = j_slice.start
        x_ = j0 + np.arange(len(z))

        dplt.push()
        fig, ax = dplt.subplots()
        ax.plot(x, y)
        ax.plot(x_, z)
        fig.tight_layout()
        dplt.show()
        dplt.pop()

    gs = GridSpec(3,3)

    fig.suptitle("Baseline Correction Inspection for " + cut_upper_folders(in_folder), fontsize=24)
    start = 0 if j_slice.start is None else j_slice.start

    topx = xr.e_curve.peak_info[peakno][1] - start
    print('topx=', topx, start+topx)

    q1 = q[i_index]
    size = data.shape[1]
    x = np.ones(size)*q1
    y = j0 + np.arange(size)

    diff_axes_info = []
    diff_ymin = None
    diff_ymax = None
    def update_diff_ylim(ax):
        ymin, ymax = ax.get_ylim()
        nonlocal diff_ymin, diff_ymax
        if diff_ymin is None or ymin < diff_ymin:
            diff_ymin = ymin
        if diff_ymax is None or ymax > diff_ymax:
            diff_ymax = ymax
    ylim_list = []
    axes_list = []

    if end_slices is None:
        end_slices = xr.e_curve.get_end_slices()

    if DEBUG:
        print("----------------- (2) end_slices=", end_slices)

    peaktop_slice = get_peaktop_slice(topx)
    bg = BaselineGuinier(data, error, q, i_index, peaktop_slice, end_slices=end_slices)
    curve0, curve1, curve2 = bg.get_scattering_curves()
    sg0, sg1, sg2 = bg.get_guinier_results()
    guinier_length0, guinier_length01, guinier_length2 = bg.get_guinier_lengths()

    yp = curve0[:,1]
    ype = curve0[:,2]
    yp_lpm1 = curve1[:,1]
    yp_lpm2 = curve2[:,1]

    try:
        if sg2.basic_quality is None or sg2.basic_quality == 0:
            gn_max = sg1.guinier_stop
        elif sg1.basic_quality is None or sg1.basic_quality == 0:
            gn_max = sg2.guinier_stop
        else:
            gn_max = max(sg1.guinier_stop, sg2.guinier_stop)
        gi = bisect_right(q, min(0.1, q[gn_max]*1.2))
    except:
        # log_exception(logger, "SimpleGuinier error: ")
        gi = bisect_right(q, 0.05)
        logger.warning("gi is set from bisect_right(q, 0.05) due to SimpleGuinier failure")
        if False:
            gslice = slice(0, gi)
            q2 = q[gslice]**2
            dplt.push()
            fig, (ax1, ax2) = dplt.subplots(ncols=2, figsize=(14, 7))
            ax1.plot(y, z)
            ax1.plot(y[topx], z[topx], 'o', color='red')
            for i, gy in enumerate([yp_lpm1, yp_lpm2]):
                ax2.plot(q2, np.log(gy[gslice]), label=str(i))
            ax2.legend()
            fig.tight_layout()
            dplt.pop()
            dplt.show()
            exit()

    gslice = slice(0, gi)
    q2 = q[gslice]**2

    def draw_scd_text(ax, scd):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = xmin*0.95 + xmax*0.05
        ty = np.power(10, np.log10(ymin)*0.9 + np.log10(ymax)*0.1)
        inc = scd - scd0
        sign  = '-' if inc < 0 else '+'
        ax.text(tx, ty, "SCD=%.2f %s %.2f" % (scd0, sign, abs(inc)), va="center", fontsize=30, alpha=0.5)

    scd_pairs = []
    ret_axes = []
    for i, xray_baseline_type in enumerate(["No Correctin", "Linear Baseline", "Integral Baseline"]):
        gs_start = i
        gs_stop = (i+1)
        ax1 = fig.add_subplot(gs[gs_start:gs_stop,0])
        ax2 = fig.add_subplot(gs[gs_start:gs_stop,1])
        ax3 = fig.add_subplot(gs[gs_start:gs_stop,2])
        # axes_list.append((ax2, ax3))
        ret_axes.append((ax1, ax2, ax3))

        for ax in [ax2]:
            ax.axes.get_xaxis().set_ticks([])

        if i == 0:
            ax1.set_title("Elution Curve", fontsize=TITLE_FONTSIZE)
            ax2.set_title("Scattering Curve (Log-Log Plot)", fontsize=TITLE_FONTSIZE)
            ax3.set_title("Scattering Curve (Guinier Plot)", fontsize=TITLE_FONTSIZE)

        ax1.plot(y, z, color='orange')

        integral = i == 2

        if i == 0:
            b = np.zeros(len(z))
        elif i == 1:
            b = compute_baseline(z, integral=False, end_slices=end_slices)
        else:
            b, convex = better_integrative_curve(z, end_slices=end_slices)

        ax1.plot(y, b, ':', color='black', linewidth=3)
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax)

        f, t = y[[peaktop_slice.start, peaktop_slice.stop]]
        p = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax1.add_patch(p)

        ax2.set_yscale("log")
        ax2.set_xscale("log")

        alpha = 0.3 if smooth_line else 1
        ax2.plot(q, yp, label='No Correction', color='blue', alpha=alpha)
        ax3.plot(q2, np.log(yp[gslice]), label='No Correction', color='blue')

        if i == 0:
            ax2.legend(loc="lower left", fontsize=LEGEND_FONTSIZE)
            sg_ = sg0

        elif i == 1:
            ax2.plot(q, yp_lpm1, label='LPM Corrected', color='red', alpha=alpha)
            ylim_list.append(ax2.get_ylim())

            if smooth_line:
                labels = ["No Correction", "Varying %"]
                for k, yp_ in enumerate([yp,  yp_lpm1]):
                    qc, ac, ec, dmax = fit_data(q, yp_, ype)
                    ax2.plot(qc, ac, color='C%d' % k, linewidth=3, label="DENSS fit-data (%s)" % labels[k])

            ax3.plot(q2, np.log(yp_lpm1[gslice]), label='LPM Corrected', color='red')
            sg_ = sg1

        else:
            for slice_ in end_slices:
                ax1.plot(y[slice_], z[slice_], 'o', color='yellow', markersize=3)

            ax2.set_yscale("log")
            ax2.set_xscale("log")

            colors = ["blue", "red"]

            ax2.plot(q, yp_lpm2, label="Integral Corrected", color=colors[1], alpha=alpha)

            if smooth_line:
                for k, yp_ in enumerate([yp, yp_lpm2]):
                    qc, ac, ec, dmax = fit_data(q, yp_, ype)
                    ax2.plot(qc, ac, color=colors[k], linewidth=3, label="DENSS fit-data (%s)" % labels[k])

            ax3.plot(q2, np.log(yp_lpm2[gslice]), label="Integral Corrected", color='red')
            sg_ = sg2

        ymin, ymax = ax3.get_ylim()
        ax3.set_ylim(ymin, ymax)

        try:
            f = q[sg_.guinier_start]**2
            t = q[sg_.guinier_stop]**2
            p = Rectangle(
                    (f, ymin),      # (x,y)
                    t - f,          # width
                    ymax - ymin,    # height
                    facecolor   = 'green',
                    alpha       = 0.2,
                    label='Guinier Region',
                )
            ax3.add_patch(p)
        except:
            pass

        ax2.legend(loc="lower left", fontsize=LEGEND_FONTSIZE)
        ax3.legend(fontsize=LEGEND_FONTSIZE)

        if show_scd:
            cd = ConcDepend(q, lpm2.data, error, ecurve_for_scd)
            scd_list = cd.compute_judge_info()
            scd2 = scd_list[peakno][1]
            scd_pairs.append((scd1, scd2))

        def set_legend_ylim(ax):
            # ax.legend(fontsize=LEGEND_FONTSIZE)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymax-4.5, ymax)

        set_legend_ylim(ax2)
        ylim_list.append(ax2.get_ylim())

    for axes in diff_axes_info:
        for ax in axes[1:]:
            ax.set_ylim(diff_ymin, diff_ymax)

    ymin, ymax = np.average(np.array(ylim_list), axis=0)
    for axes, pair in zip(axes_list, scd_pairs):
        for ax, scd in zip(axes, pair):
            ax.set_ylim(ymin, ymax)
            draw_scd_text(ax, scd)

    fig.tight_layout()
    fig.subplots_adjust(top=0.88)

    if False:

        for ax2, ax2d in diff_axes_info:
            ay0 = ax2.axes.get_position().y0
            for ax in [ax2d]:
                bbox = ax.axes.get_position()
                bbox.y1 = bbox.y0 + (ay0 - bbox.y0)*0.8
                ax.axes.set_position(bbox)

    return ret_axes

class FromToSpinBoxPair(Tk.Frame):
    def __init__(self, parent, text=None, range_=slice(0, 3), bound=(0, 10)):
        Tk.Frame.__init__(self, parent)
        if text is not None:
            label = Tk.Label(self, text=text)
            label.pack(side=Tk.LEFT)

        self.f = Tk.IntVar()
        self.f.set(range_.start)
        self.t = Tk.IntVar()
        self.t.set(range_.stop)
        self.spinboxes = []
        for k, var in enumerate([self.f, self.t]):
            f_, t_ = bound
            sbox  = Tk.Spinbox(self, textvariable=var,
                              from_=f_, to=t_, increment=1,
                              justify=Tk.CENTER, width=6)
            sbox.pack(side=Tk.LEFT)
            self.spinboxes.append(sbox)
    def get_vars(self):
        return [self.f, self.t]

    def config(self, state):
        for w in self.spinboxes:
            w.config(state=state)

class LpmInspector(Dialog):
    def __init__(self, parent, md, suppress_titles=False):
        parent.config(cursor='wait')
        parent.update()
        self.logger = logging.getLogger(__name__)
        self.parent = parent
        self.md = md
        self.j0 = md.xr.j_slice.start
        end_slices = get_setting("manual_end_slices")
        if end_slices is None:
             end_slices = md. xr.e_curve.get_end_slices()

        if DEBUG:
            print("---------------- (0): data.shape=", md.xr.data.shape, "end_slices=", end_slices)
        self.init_end_slices = end_slices
        self.primary_top_x = md.xr.e_curve.max_x - self.j0
        self.suppress_titles = suppress_titles
        self.ax = None
        self.tracing = True
        self.xray_baseline_type_auto = 0
        self.in_folder = get_setting('in_folder')
        Dialog.__init__(self, parent, "Baseline Inspector", visible=False)
        parent.config(cursor='')

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed.
        # note also that this is not only with "Cancel"
        # but also called in the normal exit with "OK".
        plt.close(self.fig)
        # print("LpmInspector: close fig")
        Dialog.cancel(self)

    def show(self):
        self._show()

    def body(self, frame):
        cframe = Tk.Frame(frame)
        cframe.pack()
        self.c1frame = c1frame = Tk.Frame(cframe)
        c1frame.pack(side=Tk.LEFT, fill=Tk.BOTH, expand=1)
        c2frame = Tk.Frame(cframe)
        c2frame.pack(side=Tk.LEFT)
        bframe = Tk.Frame(frame)
        bframe.pack(fill=Tk.X)
        adjust_frame = Tk.Frame(bframe)
        adjust_frame.pack(side=Tk.LEFT)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT, padx=30)

        self.buid_baseline_type_buttons(c1frame)

        self.fig = plt.figure(figsize=(18, 11))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, c2frame)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.build_control_widgets(adjust_frame)
        self.redraw()

    def buid_baseline_type_buttons(self, c1frame):
        self.xray_baseline_type = Tk.IntVar()
        self.xray_baseline_type.set(get_setting("xray_baseline_type"))

        for i, w in enumerate([1, 2, 2, 2]):
            c1frame.grid_rowconfigure(i, weight=w)
        space = Tk.Frame(c1frame)
        space.grid(row=0, column=0)
        rb_frame0 = Tk.Frame(c1frame)
        rb_frame0.grid(row=1, column=0)
        rb_frame1 = Tk.Frame(c1frame)
        rb_frame1.grid(row=2, column=0)
        rb_frame2 = Tk.Frame(c1frame)
        rb_frame2.grid(row=3, column=0)

        b0 = ttk.Radiobutton(rb_frame0, variable=self.xray_baseline_type, value=0)
        b0.pack()
        t0 = Tk.Label(rb_frame0, text="None", font=(None, TITLE_FONTSIZE))
        t0.pack(padx=5)
        b1 = ttk.Radiobutton(rb_frame1, variable=self.xray_baseline_type, value=1)
        b1.pack()
        t1 = Tk.Label(rb_frame1, text="Linear", font=(None, TITLE_FONTSIZE))
        t1.pack(padx=5)
        b2 = ttk.Radiobutton(rb_frame2, variable=self.xray_baseline_type, value=2)
        b2.pack()
        t2 = Tk.Label(rb_frame2, text="Integral", font=(None, TITLE_FONTSIZE))
        t2.pack(padx=5)

    def build_control_widgets(self, pframe):
        to_trace_vars = []

        self.baseline_manually = Tk.IntVar()
        self.baseline_manually.set(get_setting("baseline_manually"))
        self.baseline_manually_cb = Tk.Checkbutton(pframe, variable=self.baseline_manually, 
                                                    text="Manual Spec.")
        self.baseline_manually_cb.pack(side=Tk.LEFT, padx=20)

        label = Tk.Label(pframe, text="Peak No: ")
        label.pack(side=Tk.LEFT)
        ecurve = self.md.xr.e_curve
        self.peakno = Tk.IntVar()
        self.peakno.set(ecurve.primary_peak_no + 1)

        max_peakno = len(ecurve.peak_info)
        self.spinbox = Tk.Spinbox(pframe, textvariable=self.peakno,
                                  from_=1, to=max_peakno, increment=1,
                                  justify=Tk.CENTER, width=6)
        self.spinbox.pack(side=Tk.LEFT, padx=10)
        to_trace_vars.append(self.peakno)

        if USE_OPTINAL_BUTTONS:
            self.smooth_lines = Tk.IntVar()
            self.smooth_lines.set(0)
            toggle_cb = Tk.Checkbutton(pframe, text="show smooth lines", variable=self.smooth_lines)
            toggle_cb.pack(side=Tk.LEFT, padx=20, pady=5)
            to_trace_vars.append(self.smooth_lines)

            self.show_scd = Tk.IntVar()
            self.show_scd.set(0)
            show_scd_cb = Tk.Checkbutton(pframe, text="show SCD", variable=self.show_scd)
            show_scd_cb.pack(side=Tk.LEFT, pady=5)
            to_trace_vars.append(self.show_scd)

        tx = int(round(self.primary_top_x))
        xmax = self.md.xr.data.shape[1]

        end_ranges = self.get_init_end_ranges(xmax)
        self.sbp_list = []
        sbp = FromToSpinBoxPair(pframe, text="Integral Left Base", range_=end_ranges[0], bound=(0, tx))
        sbp.pack(side=Tk.LEFT, padx=10)
        to_trace_vars += sbp.get_vars()
        self.sbp_list.append(sbp)
        sbp = FromToSpinBoxPair(pframe, text="Integral Right Base", range_=end_ranges[1], bound=(tx, xmax))
        sbp.pack(side=Tk.LEFT, padx=10)
        to_trace_vars += sbp.get_vars()
        self.sbp_list.append(sbp)

        self.redraw_btn_blink = BlinkingFrame(pframe)
        self.redraw_btn_blink.pack(side=Tk.LEFT, padx=10, pady=5)
        self.redraw_btn = Tk.Button(self.redraw_btn_blink, text="Redraw", width=10, command=self.redraw)
        self.redraw_btn.pack()
        self.redraw_btn_blink.objects = [self.redraw_btn]

        for v in to_trace_vars:
            v.trace("w", self.blink_start)

        self.baseline_manually_tracer(xmax, init_call=True)
        self.baseline_manually.trace("w", lambda *args: self.baseline_manually_tracer(xmax))
        self.xray_baseline_type.trace("w", self.xray_baseline_type_tracer)

    def xray_baseline_type_tracer(self, *args):
        if not self.tracing:
            return

        manually = self.baseline_manually.get()
        if not manually:
            yn = MessageBox.askyesno("Manual Specification Implied",
                'Changing baseline type implies "Manual Spec."\n'
                "where the automatic baseline type control will be disabled.\n"
                "Do you really mean it?",
                parent=self.c1frame)

            if yn:
                self.baseline_manually.set(1)
            else:
                self.tracing = False
                self.xray_baseline_type.set(self.xray_baseline_type_auto)
                self.update()
                self.tracing = True

    def get_init_end_ranges(self, xmax):
        end_ranges = []
        if DEBUG:
            print("j0=", self.j0, "init_end_slices=", self.init_end_slices)
        for k, slice_ in enumerate(self.init_end_slices):
            offset = self.j0 if k == 0 else 0
            # offset is a temporary fix for ...
            start = slice_.start
            if start < 0:
                start += xmax
            stop = slice_.stop
            if stop is None:
                stop = xmax
            end_ranges.append(slice(offset+start, offset+stop))
        if DEBUG:
            print("end_ranges=", end_ranges)
        return end_ranges

    def baseline_manually_tracer(self, xmax, init_call=False):
        manually = self.baseline_manually.get()
        state = Tk.NORMAL if manually else Tk.DISABLED
        for sbp in self.sbp_list:
            sbp.config(state=state)
        self.update_selector()
        if not manually and not init_call:
            end_ranges = self.get_init_end_ranges(xmax)
            for sbp, slice_ in zip(self.sbp_list, end_ranges):
                sbp.f.set(self.j0 + slice_.start)
                sbp.t.set(self.j0 + slice_.stop)

    def blink_start(self, *args):
        self.redraw_btn_blink.start()

    def redraw(self):
        self.redraw_btn_blink.stop()
        self.config(cursor='wait')
        self.update()
        self.fig.clf()
        peakno = self.peakno.get() - 1
        kwargs = {}
        if USE_OPTINAL_BUTTONS:
            smooth_line=self.smooth_lines.get()
            show_scd = self.show_scd.get()
            kwargs = {'smooth_line':smooth_line, 'show_scd':show_scd}
        end_slices = self.get_end_slices()
        if DEBUG:
            print("----------------- (1) (must be restricted)  end_slices=", end_slices)
        ret_axes = demo_impl(self.fig, self.in_folder, self.md, j0=self.j0, peakno=peakno, end_slices=end_slices,
                                logger=self.logger, **kwargs)
        self.ax = ret_axes[2][0]
        self.update_selector()
        self.mpl_canvas.draw()
        self.config(cursor='')

    def get_end_slices(self):
        end_slices = []
        for sbp in self.sbp_list:
            # note that - self.j0 means the restricted indeces
            vals = [v.get() - self.j0 for v in sbp.get_vars()]
            end_slices.append(slice(vals[0], vals[1]))
        return end_slices

    def update_selector(self):
        if self.baseline_manually.get() and self.ax is not None:
            self.span = SpanSelector(self.ax, self.onselect, 'horizontal', useblit=True,
                        props=dict(alpha=0.5))
        else:
            self.span = None

    def onselect(self, xmin, xmax):
        i = 0 if xmin < self.primary_top_x else 1
        sbp = self.sbp_list[i]
        sbp.f.set(int(round(xmin)))
        sbp.t.set(int(round(xmax)))

    def apply(self):
        print('apply')

        manually = self.baseline_manually.get()
        set_setting('baseline_manually', manually)
        if manually:
            xray_baseline_type = self.xray_baseline_type.get()
            end_slices = self.get_end_slices()
        else:
            xray_baseline_type = None
            end_slices = None
        set_setting('xray_baseline_type', xray_baseline_type)
        set_setting('manual_end_slices', end_slices)

    def save_the_figure(self, folder=None, file=None):
        import os
        from molass_legacy._MOLASS.SerialSettings import get_setting
        if folder is None:
            folder = get_setting('analysis_folder')
        if file is None:
            file = get_setting('analysis_name')
        path = os.path.join(folder, file)
        self.fig.savefig(path)



"""
    MpInverseBasic.py
"""
import numpy as np
import matplotlib.pyplot as plt

x = np.linspace(1, 24, 100)

fig, ax = plt.subplots()
ax.set_title(r"$M=P \cdot C$ Metaphor", fontsize=20)
ax.set_aspect("equal")
ax.set_xlabel("C", fontsize=16)
ax.set_ylabel("P", fontsize=16)

ax.plot(x, 24/x, label=r"$P \cdot C=24$")
ax.plot(3, 8, "o", color="red")
ax.plot(5, 8, "o")
ax.plot(5, 4.8, "o")

ax.annotate(r"$(C,P)=(3,8)$", xy=(3,8), xytext=(4,15), arrowprops=dict( arrowstyle="-", color="gray" ), fontsize=16 )
ax.annotate("", xy=(5,8), xytext=(3,8), arrowprops=dict( arrowstyle="->", color="gray" ))
ax.annotate("", xy=(5,4.8), xytext=(5,8), arrowprops=dict( arrowstyle="->", color="gray" ))

ax.legend(fontsize=16)
fig.tight_layout()
plt.show()



"""
    Models.Stochastic.OligoporeMoments.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize, basinhopping
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments, to_moderate_props
from molass_legacy.Models.Stochastic.RgReliability import determine_unreliables
from molass_legacy.Models.Stochastic.ParamLimits import USE_K, M1_WEIGHT, M2_WEIGHT, PORESIZE_BOUNDS, BASINHOPPING_SCALE, N0_INIT, OligoporeParamsScaler
from molass_legacy.Models.Stochastic.OligoporeImpl import USE_DISPERSIVE, oligopore_pdf

M1_WEIGHT = 9
M2_WEIGHT = 1
PENALTY_SCALE = 1e6
PORESIZE_LIMIT = 1000

def sec_oligopore_pdf_proof(lrf_src):
    from SecTheory.BasicModels import robust_single_pore_pdf

    N, T, t0, me, mp, poresize = 1000, 0.5, 50, 1.5, 1.5, 100

    pszs = np.ones(3) * poresize
    pszp = np.ones(3)/3

    rg = 30
    rhov = rg/pszs
    ni_ = N * (1 - rhov)**me
    ti_ = T * (1 - rhov)**mp
    x = np.arange(300)
    y = robust_single_pore_pdf(x - t0, ni_[0], ti_[0])
    N0 = N0_INIT if USE_DISPERSIVE else None

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Proof of sec_oligopore_pdf compared to robust_single_pore_pdf")
        ax.plot(x, y)
        nt_pairs = [(ni_, ti_) for ni_, ti_ in zip(ni_, ti_)]
        py = oligopore_pdf(x - t0, nt_pairs, pszp, 0, N0=N0)
        ax.plot(x, py)
        py = oligopore_pdf(x, nt_pairs, pszp, t0, N0=N0)
        ax.plot(x, py)        
        ax.axvline(t0, color="red")
        # ax.axvline(x0, color="gray")
        fig.tight_layout()
        ret = plt.show()
    return ret

def study_oligopore_moments_impl(lrf_src,
                                 keep_num_components=False,
                                 want_num_components=None,
                                 trust_max_num=None,
                                 return_guess_params=False,
                                 debug=False):
    if debug:
        import Models.Stochastic.OligoporeUtils as module
        reload(module)
    from molass_legacy.Models.Stochastic.OligoporeUtils import guess_oligopore_colparams, plot_oligopore_conformance

    print('study_oligopore_moments_impl')

    if False:
        ret = sec_oligopore_pdf_proof(lrf_src)
        if not ret:
            return

    all_peaks = lrf_src.xr_peaks
    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(keep_num_components=keep_num_components,
                                                                                              want_num_components=want_num_components,
                                                                                              debug=False)
    print("peak_rgs=", peak_rgs)

    peaks = all_peaks[indeces]
    num_peaks = len(peaks)
    egh_moments_list = compute_egh_moments(peaks)
    m_props = to_moderate_props(props)

    # num_pszs = 3        # i.e., Tripore
    num_pszs = 2        # i.e., Dipore
    me = 1.5
    mp = 1.5

    unreliable_indeces = determine_unreliables(peak_rgs, qualities, props, trust_max_num=trust_max_num)
    nur = len(unreliable_indeces)

    use_basinhopping = True         # it seems that global optimization is required for this problem
    params_scaler = OligoporeParamsScaler(num_pszs, egh_moments_list, 0, peak_rgs, unreliable_indeces, dispersive=USE_DISPERSIVE)

    # scales to normalize the parameters for basinhopping
    # param_scales = np.array([100, 0.1, 50, 100] + [10]*num_pszs + [0.1]*num_pszs + [10]*nur)  

    temp_rgs = np.asarray(peak_rgs).copy()
    egh_moments_array = np.asarray(egh_moments_list)

    sep = 0
    slice1 = slice(0, 3)
    sep += 3
    slice2 = slice(sep, sep + num_pszs)
    sep += num_pszs
    slice3 = slice(sep, sep + num_pszs-1)
    sep += num_pszs-1
    slice4 = slice(sep, sep + nur)
    sep += nur
    N0_index = -1
    def split_params(p):
        N_, KT_, x0_ = p[slice1]
        if USE_K:
            T_ = KT_/N_
        else:
            T_ = KT_
        props = p[slice3]
        props_ = np.concatenate([props, [1 - np.sum(props)]])
        N0 = p[N0_index] if USE_DISPERSIVE else None
        return (N_, T_, x0_), p[slice2], props_, p[slice4], N0

    def plp_column_flex_objective(p, debug=False):
        if use_basinhopping:
            p = params_scaler.scale_back(p)
        (N_, T_, x0_), pszv_, pszp_, p_rgs, N0_ = split_params(p)
        temp_rgs[unreliable_indeces] = p_rgs
        
        psz_order_penalty = min(0, np.min(np.diff(pszv_)))**2 *PENALTY_SCALE    # poresizes should be in ascending order
        negative_penalty = min(0, np.min(pszp_))**2 *PENALTY_SCALE

        dev_list = []
        for k, (M, rg) in enumerate(zip(egh_moments_list, temp_rgs)):
            rhov = rg/pszv_
            rhov[rhov > 1] = 1
            M1_ = x0_ + N_ * T_ * np.sum(pszp_ * (1 - rhov)**(me + mp))
            if N0_ is None:
                M2_ = 2 * N_ * T_**2 * np.sum(pszp_ * (1 - rhov)**(me + 2*mp))
            else:
                M2_ = np.sum( 2 * N_ * T_**2 * pszp_ * (1 - rhov)**(me + 2*mp ) + (pszp_ * ((1 - rhov)**(me + mp) + x0_))**2/N0_ )
            dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)
            if debug:
                print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))
        if debug:
            print("pszv_, pszp_=", pszv_, pszp_)
            print("psz_order_penalty=", psz_order_penalty)
            print("negative_penalty=", negative_penalty)

        return np.sum(np.asarray(dev_list) * m_props) + psz_order_penalty + negative_penalty

    if USE_K:
        kt_init = np.sqrt(np.sum(egh_moments_array[:,1])) * 30
    else:
        kt_init = 0.5
    t0_init = np.min([M[0] - np.sqrt(M[1])*20 for M in egh_moments_list])
    if num_pszs == 2:
        poresizes_init = [80, 200]
    else:
        poresizes_init = [70, 100, 200]
    assert len(poresizes_init) == num_pszs

    colparams = [1000, kt_init, t0_init] + poresizes_init + list(np.ones(num_pszs-1)/num_pszs)

    plot_info = Struct(x=lrf_src.xr_x, y=lrf_src.xr_y, model=lrf_src.model, peaks=peaks)
    if num_peaks > 2:
        ret = guess_oligopore_colparams(num_pszs, peak_rgs, qualities, egh_moments_list, colparams, m_props, debug=debug, plot_info=plot_info)
        if ret is None:
            return
        if return_guess_params:
            return ret
        N, K, t0, pszs, pszp = ret
        colparams = np.concatenate([[N, K, t0], pszs, pszp[:-1]])   #  remember that pszp has an extra value to make it sum up to unity
        print("colparams=", colparams)

    init_params_list = [colparams,
                        peak_rgs[unreliable_indeces],
                        [N0_INIT] if USE_DISPERSIVE else [],
                        ]

    init_params = np.concatenate(init_params_list)

    if use_basinhopping:
        init_params_ = params_scaler.scale(init_params)
        bounds_ = [(0, BASINHOPPING_SCALE)] * len(init_params)
        plp_column_flex_objective(init_params_, debug=True)
        res = basinhopping(plp_column_flex_objective, init_params_, niter=100,
                           minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds_))
        plp_column_flex_objective(res.x, debug=True)
        res.x = params_scaler.scale_back(res.x)     # do not forget to scale back the result
    else:
        bounds = params_scaler.get_bounds()
        res = minimize(plp_column_flex_objective, init_params, bounds=bounds, method='Nelder-Mead')

    if debug:
        print("mnp_column_objective: res.x=", res.x)
    
    x = lrf_src.xr_x
    y = lrf_src.xr_y

    (N, T, x0), pszv, pszp, p_rgs, N0 = split_params(res.x)
    temp_rgs[unreliable_indeces] = p_rgs
    abort = False

    def plp_scales_objective(p, return_cy_list=False, title=None):
        debug = title is not None
        # x0_ = p[0]
        scales = p
        cy_list = []
        if debug:
            moments = []
        for k, (rg, scale) in enumerate(zip(temp_rgs, scales)):
            rhov = rg/pszv
            rhov[rhov > 1] = 1
            if debug:
                M1_ = x0 + N * T * np.sum(pszp * (1 - rhov)**(me + mp))
                if N0 is None:
                    M2_ = 2 * N * T**2 * np.sum(pszp * (1 - rhov)**(me + 2*mp ))
                else:
                    M2_ = np.sum( 2 * N * T**2 * pszp * (1 - rhov)**(me + 2*mp ) + (pszp * ((1 - rhov)**(me + mp) + x0))**2/N0 )
                moments.append([M1_, M2_])
            ni_ = N * (1 - rhov)**me
            ti_ = T * (1 - rhov)**mp
            nt_pairs = [(ni_, ti_) for ni_, ti_ in zip(ni_, ti_)]
            cy = scale * oligopore_pdf(x, nt_pairs, pszp, x0, N0=N0)
            cy_list.append(cy)
        if return_cy_list:
            return cy_list
        ty = np.sum(cy_list, axis=0)
        dev = np.sum((ty - y)**2)
        if debug:
            def plot_details():
                import Models.Stochastic.OligoporeUtils
                reload(Models.Stochastic.OligoporeUtils)
                from molass_legacy.Models.Stochastic.OligoporeUtils import plot_oligopore_moments_state
                params_list = [(N, T, x0), pszv, pszp, scales]      # not including me, mp here
                if N0 is not None:
                    params_list += [[N0]]
                olgp_params = np.concatenate(params_list)
                ret = plot_oligopore_moments_state(x, y, cy_list, ty, egh_moments_list, moments, split_params, olgp_params, temp_rgs, params_scaler, title=title)

            with plt.Dp(button_spec=['OK', 'Cancel'],
                        extra_button_specs=[("Plot Details", plot_details)]
                        ):
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y, color='orange', label='data')
                for k, cy in enumerate(cy_list):
                    ax.plot(x, cy, ":", label='component-%d' % k)
                ax.plot(x, ty, ":", color='red', label='model total')
                fig.tight_layout()
                ret = plt.show()
            if not ret:
                nonlocal abort
                abort = True
        return dev
    
    cy_list = plp_scales_objective(np.ones(num_peaks), return_cy_list=True)
    init_scales = []
    for cy, h in zip(cy_list, peaks[:,0]):
        init_scales.append(h/np.max(cy))
    init_params = np.array(init_scales)
    if debug:
        plp_scales_objective(init_params, title="before MNP with unreliables %s minimize" % unreliable_indeces)
        if abort:
            return
    # bounds = [(-500, 500)] + [(0, 10)]*num_peaks
    bounds = [(0, 10)]*num_peaks
    if True:
        res = minimize(plp_scales_objective, init_params, bounds=bounds, method='Nelder-Mead')
    else:
        res = basinhopping(plp_scales_objective, init_params, niter=100,
                           minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds))        
    if debug:
        plp_scales_objective(res.x, title="after MNP with unreliables %s minimize" % unreliable_indeces)
        if abort:
            return

    # x0 = res.x[0]
    scales = res.x
    params_list = [(N, T, x0, me, mp), pszv, pszp, scales]
    if N0 is not None:
        params_list += [[N0]]

    oligopore_params = np.concatenate(params_list)
    if debug:
        K = N*T
        ret = plot_oligopore_conformance(N, K, t0, pszs, pszp, plot_info, rg_info=(temp_rgs, peak_trs, True))
        if not ret:
            return

    return oligopore_params, temp_rgs, unreliable_indeces


# -*- coding: utf-8 -*-
"""

    ファイル名：   CommandController.py

    処理内容：

        コマンドモードの実行制御

"""
from __future__ import division, print_function, unicode_literals
import sys
import os
sys.path.append( os.path.dirname( os.path.abspath( __file__ ) ) + '/ExecutionWindow/bin' )

import shutil
import logging
from OurException       import OurException
from SynthesizerSettings    import set_setting, set_mask
from molass_legacy.KekLib.BasicUtils         import mkdirs_with_retry, exe_name
from PilatusUtils       import get_data_info
from ImageSynthesizer   import ImageSynthesizer
from Preferences        import set_preference
from OurMock            import MagicMock
import ChangeableLogger

class Controller:

    def __init__( self, opts ):
        self.opts   = opts

        if opts.in_folder == None:
            raise OurException( '-i IN_FOLDER argument is required!' )

        if not os.path.exists( opts.in_folder ):
            raise OurException( '%s does not exist!' % ( opts.in_folder ) )

        if not os.path.isdir( opts.in_folder ):
            raise OurException( '%s is not a folder!' % ( opts.in_folder ) )

        self.new_out_folder_created = False
        if opts.out_folder == None:
            new_file_ok = False
            out_folder = opts.in_folder + '/Synthesized'

            if not opts.autonum_folders:
                if os.path.exists( out_folder ):
                    shutil.rmtree( out_folder )

            for i in range( 1, 100 ):
                if os.path.exists( out_folder ):
                    out_folder = opts.in_folder + '/Synthesized(%d)' % ( i )
                else:
                    new_file_ok = True
                    break
            assert( new_file_ok )
            mkdirs_with_retry( out_folder )
            opts.out_folder = out_folder
            self.new_out_folder_created = True

        if not os.path.exists( opts.out_folder ):
            raise OurException( '%s does not exist!' % ( opts.out_folder ) )

        if not os.path.isdir( opts.out_folder ):
            raise OurException( '%s is not a folder!' % ( opts.out_folder ) )

        set_setting( 'in_folder', opts.in_folder )
        set_setting( 'adj_folder', None )
        set_setting( 'syn_folder', opts.out_folder )

    def execute( self ):
        counter_id      = 'None'
        pilatus_counter = None
        in_folder       = self.opts.in_folder
        adj_folder      = self.opts.adj_folder
        out_folder      = self.opts.out_folder

        self.logfile_path   = '%s/%s.log' % ( out_folder, exe_name() )
        app_logger  = ChangeableLogger.Logger( self.logfile_path )
        logger      = logging.getLogger( __name__ )

        if self.new_out_folder_created:
            logger.info( "New folder '%s' has been created." % ( out_folder ) )

        try:
            log_file, mask_file, data_array, pilatus_counter = get_data_info( in_folder, adj_folder, out_folder, pilatus_counter, counter_id )

            set_mask( os.path.join( in_folder, mask_file) )

            synthesizer = ImageSynthesizer()
            synthesizer.set_setting_info( 3 )
            synthesizer.exec_array = data_array
            mockconnector   = MagicMock()
            mockprogress    = MagicMock()
            synthesizer.exec_syntheses( mockconnector, mockprogress )
        except:
            logging.exception( 'Unexpected error' )



"""
    UvBaseSolverInfo.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import curve_fit
import molass_legacy.KekLib.DebugPlot as plt
from .Sigmoid import ex_sigmoid
from molass_legacy.Peaks.EghSupples import d_egh

TRY_SPECIAL_SIGMOID = False

def make_info_lists(fc, debug=False):
    egh_param_list = fc.a_curve.get_model_param_list()
    sigmoid_param_list = guess_better_sigmoid_params(fc, debug=debug)
    d_egh_param_list = guess_distortion_params(fc, egh_param_list, sigmoid_param_list, debug=debug)
    return egh_param_list, sigmoid_param_list, d_egh_param_list

def guess_better_sigmoid_params(fc, debug=False):
    sigmoid_param_list = [popt for popt in fc.popts if popt is not None]

    if len(sigmoid_param_list) == 0 and TRY_SPECIAL_SIGMOID:
        # as in 20231214/PgpApo
        try:
            if debug:
                from importlib import reload
                import Trimming.SpecialSigmoid
                reload(Trimming.SpecialSigmoid)
            from .SpecialSigmoid import guess_special_sigmoid_params
            sigmoid_param_list = guess_special_sigmoid_params(fc, debug=debug)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "guess_special_sigmoid_params failed: ")

    num_sigmoids = len(sigmoid_param_list)
    assert num_sigmoids > 0                 # num_sigmoids == 0 case must fail
    num_sigmoid_params = 6

    x = fc.a_curve2.x
    y = fc.a_curve2.y

    def sigmoid_chain_func(x, *p):
        cy = np.zeros(len(x))
        start = 0
        for k in range(num_sigmoids):
            stop = start+num_sigmoid_params
            cy += ex_sigmoid(x, *p[start:stop])
            start = stop
        return cy

    p0 = np.array(sigmoid_param_list).flatten()

    if debug:
        print("sigmoid_param_list=", sigmoid_param_list)
        print("p0=", p0)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_better_sigmoid_params before")
            ax.plot(x, y)
            ax.plot(x, sigmoid_chain_func(x, *p0))
            plt.show()

    popt, pcov = curve_fit(sigmoid_chain_func, x, y, p0)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_better_sigmoid_params after")
            ax.plot(x, y)
            ax.plot(x, sigmoid_chain_func(x, *popt))
            plt.show()

    ret_list = []
    start = 0
    for k in range(num_sigmoids):
        stop = start + num_sigmoid_params
        ret_list.append(popt[start:stop].copy())
        start = stop
    return ret_list

def guess_distortion_params(fc, egh_param_list, sigmoid_param_list, debug=False):
    num_peaks = len(egh_param_list)
    num_egh_params = 4

    x = fc.a_curve2.x
    y = fc.a_curve2.y

    y_ = y.copy()
    for params in sigmoid_param_list:
        y_ -= ex_sigmoid(x, *params)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_distortion_params before")
            ax.plot(x, y)
            ax.plot(x, y_)
            plt.show()

    egh_param_list_ = [prm.copy()for prm in egh_param_list]
    def distortion_curve_func(x, *p):
        cy = np.zeros(len(x))
        for k in range(num_peaks):
            params = egh_param_list_[k]
            params[0] = p[k]
            cy += d_egh(x, *params)
        return cy

    p0_list = []
    for params in egh_param_list:
        # change H
        p0_list.append(-params[0]*0.5)      # it should be of opposite sign

    p0 = np.array(p0_list)
    popt, pcov = curve_fit(distortion_curve_func, x, y_, p0)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_distortion_params after")
            ax.plot(x, y)
            ax.plot(x, y_)
            ax.plot(x, distortion_curve_func(x, *popt))
            plt.show()

    return popt



# coding: utf-8
"""
    NoseUtils.py

    Copyright (c) 2019-2022, Masatsuyo Takahashi, KEK-PF
"""

import sys
import re
import inspect
import traceback
import logging
import numpy as np

filepath_re = re.compile(r'File "(.+)"')
eq_call_re = re.compile(r'^(\s+(eq_|gt_)\(\s*[^,]+\s*,\s*)(.*\S)(\s*\)\s*)$')
assert_almost_equal_re = re.compile(r"^(\s+assert_almost_equal\([^,]+,\s*)([\d\.e\+\-]+)(.+)$")

success_result_dict = {}

def eq_(lhs, rhs):
    lineno = traceback.extract_stack()[2].lineno
    # print(lineno)
    # print(lhs, rhs)
    success_result_dict[lineno] = ('eq_', lhs, rhs)

def gt_(lhs, rhs):
    lineno = traceback.extract_stack()[2].lineno
    # print(lineno)
    # print(lhs, rhs)
    success_result_dict[lineno] = ('gt_', lhs, rhs)

def assert_almost_equal(lhs, rhs, n=None):
    lineno = traceback.extract_stack()[2].lineno
    success_result_dict[lineno] = ('assert_almost_equal', lhs, rhs)

def str_for_source_code(value):
    if isinstance(value, (str, Exception)):
        str_value = str(value)
        if str_value.find("'") >= 0:
            q = '"'
        else:
            q = "'"

        return q + str_value + q
    else:
        return str(value)

def write_the_successful_script(newline='\n', kwargs={}):
    # get the caller module
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])

    func_list = []
    for name, func in inspect.getmembers(module):
        if name.find('test') == 0:
            print(name)
            num = inspect.getsourcelines(func)[-1]
            # print(first_line_info)
            # num = first_line_info[1]
            func_list.append( (num, name, func) )

    for num, name, func in sorted(func_list):
        print([num], "testing", name)
        try:
            func(**kwargs)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print('%d %s call failed:\n' % (num, name), etb)

    # print('success_result_dict=', success_result_dict)

    in_filepath = module.__file__
    out_filepath = in_filepath.replace('.py', '-success.py')

    out_lines = []
    with open(in_filepath) as fh:
        for n, line in enumerate(fh.readlines(), start=1):
            success_rec = success_result_dict.get(n)
            if success_rec is None:
                out_line = line
            else:
                method, lhs, rhs = success_rec
                # assert method in ['eq_', 'gt_']
                if method == 'eq_':
                    out_line = re.sub(eq_call_re, lambda m: m.group(1) + str_for_source_code(lhs) + m.group(4), line)
                elif method == 'gt_':
                    rhs_ = min(rhs, lhs)
                    delta = np.power(10, np.floor(np.log10(rhs_)-2))
                    less_valuse = '%.3g' % (rhs_ - delta)
                    out_line = re.sub(eq_call_re, lambda m: m.group(1) + less_valuse + m.group(4), line)
                elif method == 'assert_almost_equal':
                    lhs_str = "%g" % lhs
                    out_line = re.sub(assert_almost_equal_re, lambda m: m.group(1) + lhs_str + m.group(3), line)
                else:
                    assert False
            out_lines.append(out_line)

    with open(out_filepath, 'w', newline=newline) as fh:
        fh.write( ''.join(out_lines) )

"""
    The following are dummy objects to replace ChangeableLogger instances
"""

class Logger(logging.Logger):
    def __init__(self, path):
        logging.Logger.__init__(self, __name__)

def create_our_logger(path):
    return Logger(path)



"""
    GuiParts.ColumnTypeSelector.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, ttk
from molass_legacy.Experiment.ColumnTypes import get_productnames, get_columntype_from_name
from molass_legacy.Experiment.DataUtils import get_columntype

class ColumnTypeSelector(Tk.Frame):
    def __init__(self, parent, column_name_var=None, excl_limit_var=None):
        Tk.Frame.__init__(self, parent)
        columntype = get_columntype()
        column_name = columntype.name
        if column_name_var is None:
            column_name_var = Tk.StringVar()
        self.column_name = column_name_var
        self.column_name.set(column_name)
        if excl_limit_var is None:
            excl_limit_var = Tk.DoubleVar()
        self.excl_limit = excl_limit_var
        self.excl_limit.set(columntype.excl_limit)
        self.column_products = ttk.Combobox(self, textvariable=self.column_name,
                                            width=35, justify=Tk.CENTER, state="readonly")
        self.column_products.pack(side=Tk.LEFT)
        self.column_products['values'] = get_productnames()
        self.column_name.trace("w", self.column_name_tracer)

    def config(self, **kwargs):
        self.column_products.config(**kwargs)

    def column_name_tracer(self, *args):
        columntype = get_columntype_from_name(self.column_name.get())
        self.excl_limit.set(columntype.excl_limit)


"""
    SMC.ModifiedCustomDist.py
    
    Copyright (c) 2024, SAXS Team, KEK-PF   
"""
import numpy as np
import pymc as pm

class ModifiedCustomDist(pm.CustomDist):
    """
    from the original CustomDist class documentation
    https://github.com/pymc-devs/pymc/blob/main/pymc/distributions/custom.py

        random : Optional[Callable]
            A callable that can be used to generate random draws from the distribution

            It must have the following signature: ``random(*dist_params, rng=None, size=None)``.
            The numerical distribution parameters are passed as positional arguments in the
            same order as they are supplied when the ``CustomDist`` is constructed.
            The keyword arguments are ``rng``, which will provide the random variable's
            associated :py:class:`~numpy.random.Generator`, and ``size``, that will represent
            the desired size of the random draw. If ``None``, a ``NotImplemented``
            error will be raised when trying to draw random samples from the distribution's
            prior or posterior predictive.
    """
    def __new__(cls, name, *args, **kwargs):
        def rand_cunstom(params, rng=None, size=None):

            return 
        kwargs['random'] = rand_cunstom
        return pm.CustomDist.__new__(cls, name, *args, **kwargs)


"""

    TrimmingDebugUtils.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Elution.CurveUtils import simple_plot
from .TrimmingInfo import get_trimming_ends_from_list

def trimming_debug_plot(title, sd, uv_restrict_list, xr_restrict_list ):
    a_curve = sd.absorbance.a_curve
    D, E, qv, e_curve = sd.get_xr_data_separate_ly()
    uv_restrict_list
    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
        fig.suptitle(title)

        uv_e_ends, uv_a_ends = get_trimming_ends_from_list(uv_restrict_list)
        simple_plot(ax1, a_curve, color="blue")
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax)
        for j in uv_e_ends:
            ax1.plot([j, j], [ymin, ymax], color="yellow")

        xr_e_ends, xr_a_ends = get_trimming_ends_from_list(xr_restrict_list)
        simple_plot(ax2, e_curve, color="orange")
        ymin, ymax = ax2.get_ylim()
        ax2.set_ylim(ymin, ymax)
        for j in xr_e_ends:
            ax2.plot([j, j], [ymin, ymax], color="yellow")

        ax3.set_yscale("log")
        k = e_curve.primary_peak_i
        y = np.average(D[:,k-5:k+6], axis=1)
        ax3.plot(qv, y)
        ymin, ymax = ax3.get_ylim()
        ax3.set_ylim(ymin, ymax)
        for i in xr_a_ends:
            q = qv[i]
            ax3.plot([q, q], [ymin, ymax], color="yellow")
        fig.tight_layout()
        plt.show()

def plot_restrict_lines(ax, e_restrict, **kwargs):
    if e_restrict is None:
        return

    flag, start, stop, size = e_restrict
    if not flag:
        return

    ymin, ymax = ax.get_ylim()
    for p in [start, stop]:
        if p is None:
            continue

        ax.plot([p, p], [ymin, ymax], ':', **kwargs)

def plot_side_sigma_points(ax, restrict_info, curve, emg_peaks, sigma_point_ratio, **kwargs):
    if not restrict_info.flag:
        return

    x = curve.x
    for k, peak in enumerate(emg_peaks):
        print([k], peak.opt_params)
        y = peak.get_model_y(x)
        ax.plot(x, y, ':', linewidth=3, color='green', label='approximate model')

    sigma_min = restrict_info.start
    sigma_max = restrict_info.stop
    ymin, ymax = ax.get_ylim()
    print('plot_side_sigma_points:', sigma_min, sigma_max)
    for p in [sigma_min, sigma_max]:
        ax.plot([p, p], [ymin, ymax], label='%g-sigma boundary' % sigma_point_ratio, **kwargs)

def trimming_result_plot_impl(fig, axes, sd, curves, emg_peaks_list, old_e_restricts, ret_info, sigma_point_ratio, mpeaks, alt_info=None):
    from matplotlib.patches import Rectangle
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

    in_folder = get_in_folder()
    shapes = sd.conc_array.shape, tuple(reversed(sd.intensity_array.shape[0:2]))
    vectors = sd.lvector, sd.qvector
    uv_curve, xr_curve = curves
    uv_e_restrict, xr_e_restrict = old_e_restricts
    new_e_restricts = [ret_info[0][0], ret_info[1][0]]

    if alt_info is None:
        alt_text = ""
        shapes_ = shapes
        vectors_ = vectors
        ret_info_ = ret_info
        vnames = ("w", "q")
    else:
        alt_text = " with altering info"
        shapes_ = shapes[1], shapes[1]
        vectors_ = vectors[1], vectors[1]
        ret_info_ = [alt_info[1], ret_info[1]]
        vnames = ("q", "q")

    fig.suptitle("Trimming Info at a glance for %s%s" % (in_folder, alt_text), fontsize=20)

    for ax, title in zip(axes[0,:], ["Range Rectangle", "Elution Range", "Opposite Range"]):
        ax.set_title(title, fontsize=16)

    ret_twinxes = []
    for ax, shape, vector, info_list, vname in zip(axes[:,0], shapes_, vectors_, ret_info_, vnames):
        ax.invert_yaxis()
        e_info, a_info = info_list
        print("info_list=", info_list)
        e_max = shape[1] - 1
        a_max = shape[0] - 1
        ax.plot([0, e_max, e_max, 0, 0],
                [0, 0, a_max, a_max, 0])

        e_stt = 0 if e_info is None else e_info.start
        e_end = e_max if e_info is None else e_info.end
        a_stt = 0 if a_info is None else a_info.start
        a_end = a_max if a_info is None else a_info.end

        range_width = e_end - e_stt
        range_height = a_end - a_stt
        rect = Rectangle(
                (e_stt, a_stt), # (x,y)
                range_width,    # width
                range_height,   # height
                facecolor   = 'yellow',
                alpha       = 0.3,
            )
        ax.add_patch(rect)

        whole_width = e_max
        whole_height = a_max
        is_narrow = range_width/whole_width < 0.7

        ty = (a_stt + a_end)/2
        for k, j in enumerate([e_stt, e_end]):
            dx = whole_width * (0.5 - k) * 0.2
            if is_narrow:
                dy = whole_height * (0.5 - k) * 0.2
            else:
                dy = 0
            ax.annotate("%d" % j, xy=(j, ty+dy), xytext=(j+dx, ty+dy), ha="center", va="center", arrowprops=dict(arrowstyle="->", color='k'))

        tx = (e_stt + e_end)/2
        for k, i in enumerate([a_stt, a_end]):
            dy = whole_height * (0.5 - k) * 0.25
            ax.annotate("%s[%d]=%.3g" % (vname, i, vector[i]), xy=(tx, i), xytext=(tx, i+dy), ha="center", va="center", arrowprops=dict(arrowstyle="->", color='k'))

        ax.text(tx, ty, "shape=%s" % str(shape), ha="center", va="center")

        axt = ax.twinx()
        axt.invert_yaxis()
        axt.grid(False)
        axt.plot([0, e_max, e_max, 0, 0],
                vector[[0, 0, a_max, a_max, 0]], alpha=0)
        ret_twinxes.append(axt)

    axes_ = axes[:,1]
    ax1, ax2 = axes_
    if alt_info is None:
        simple_plot(ax1, uv_curve, color='blue', legend=False)
    else:
        simple_plot(ax1, xr_curve, color='orange', legend=False)
    simple_plot(ax2, xr_curve, color='orange', legend=False)

    for ax in axes_:
        ax.set_ylim(ax.get_ylim())

    if False:
        plot_restrict_lines(ax1, uv_e_restrict, color='gray')
        plot_restrict_lines(ax2, xr_e_restrict, color='gray')

    if alt_info is None:
        new_e_restricts_ = new_e_restricts
        curves_ = curves
        emg_peaks_list_ = emg_peaks_list
    else:
        new_e_restricts_  = [alt_info[1][0], ret_info[1][0]]
        curves_ = curves[1], curves[1]
        emg_peaks_list_ = [emg_peaks_list[1], emg_peaks_list[1]]

    for ax, restrict_info, curve, emg_peaks in zip(axes_, new_e_restricts_, curves_, emg_peaks_list_):
        plot_side_sigma_points(ax, restrict_info, curve, emg_peaks, sigma_point_ratio, color='yellow', lw=3)
        ax.legend()

    from molass_legacy.DataStructure.UvSpecCurve import UvSpecCurve
    from molass_legacy.DataStructure.XrSpecCurve import XrSpecCurve
    uv_scurve = UvSpecCurve(sd, mpeaks=mpeaks)
    xr_scurve = XrSpecCurve(sd)

    if alt_info is None:
        scurve0_tuple = (axes[0,2], None,  vectors[0], uv_scurve.get_y(), None, ret_info[0][1], shapes[0])
    else:
        scurve0_tuple = (axes[0,2], "log", vectors[1], xr_scurve.get_y(), None, alt_info[1][1], shapes[1])

    for ax, yscale, x, y, colorm, a_info_, shape in [
        scurve0_tuple,
        (axes[1,2], "log", vectors[1], xr_scurve.get_y(), None, ret_info[1][1], shapes[1]),
        ]:
        if yscale is not None:
            ax.set_yscale(yscale)
        ax.plot(x, y)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        if a_info_ is None:
            start = 0
            end = shape[0] - 1
        else:
            start = a_info_.start
            end = a_info_.end
        for j in [start, end]:
            x_ = x[j]
            ax.plot([x_, x_], [ymin, ymax], color="yellow", lw=3)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)

    return ret_twinxes

def trimming_result_plot(sd, curves, emg_peaks_list, old_e_restricts, ret_info, sigma_point_ratio, mpeaks, figfile=None):
    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18,9))
        trimming_result_plot_impl(fig, axes, sd, curves, emg_peaks_list, old_e_restricts, ret_info, sigma_point_ratio, mpeaks)
        if figfile is None:
            plt.show()
        else:
            plt.show(block=False, pause=1)
            fig.savefig(figfile)



# coding: utf-8
"""
    SecTools.CorMap.AngularUnit.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np

def angstrom_qv(qv):
    if 3 < qv[-1] and qv[-1] < 8:
        qv = qv * 0.1
    return qv



"""
    Selective.PropOptimizerSTC.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from time import sleep, time
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Models.Stochastic.Tripore import guess_multiple_impl, tripore_func
from molass_legacy.KekLib.OurTkinter import Tk
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from .PropOptimizerImpl import compute_range_rgs, PROP_MIN_VALUE, RangeRgComputer

VERY_SMALL_RG = 1

def guess_init_params(x, y):
    num_peaks = 3
    params_array = np.array(guess_multiple_impl(x, y, num_peaks))
    return params_array

def compute_props(x, params_array):
    areas = []
    for params in params_array:
        cy = tripore_func(x, *params)
        areas.append(np.sum(cy))
    props = np.array(areas)/np.sum(areas)
    return props

def compute_cy_list(x, params_array):
    cy_list = []
    for params in params_array:
        cy = tripore_func(x, *params)
        cy_list.append(cy)
    return cy_list

def optimize_to_props(x, y, peaks, props, use_basinhopping=False, debug=False):
    assert len(peaks) == len(props)
    peaks = np.asarray(peaks)
    props = np.asarray(props)

    def objective(p, debug_info=None):
        ty = np.zeros(len(y))
        p_ = p.reshape(peaks.shape)
        areas = []
        try:
            cy_list = []
            for params in p_:
                cy = tripore_func(x, *params)
                cy_list.append(cy)
                areas.append(np.sum(cy))
                ty += cy
            props_ = np.array(areas)/np.sum(areas)
            area_dev = np.sum((ty - y)**2)
            prop_dev = max(PROP_MIN_VALUE, np.sum((props_ - props)**2))
            fv = np.log(area_dev) + np.log(prop_dev) - np.log(np.std(p_[:,1]))
            if debug_info is not None:
                print("props=", props)
                title, = debug_info
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title(title)
                    ax.plot(x, y)
                    for cy in cy_list:
                        ax.plot(x, cy, ":")
                    ax.plot(x, ty, ":", color="red")
                    fig.tight_layout()
                    plt.show()
        except:
            fv = np.inf
        return fv

    init_params = peaks.flatten()
    if debug:
        objective(init_params, debug_info=("optimize_to_props: before minimize", ))
    if use_basinhopping:
        ret = basinhopping(objective, init_params, minimizer_kwargs=dict(method="Nelder-Mead"))
    else:
        ret = minimize(objective, init_params, method="Nelder-Mead")
    if debug:
        objective(ret.x, debug_info=("optimize_to_props: after minimize", ))
    return ret

def compute_optimal_proportion(progress_queue, job_args):
    for i in range(10):
        print([i])
        sleep(1)
        progress_queue.put([i])
    progress_queue.put([-1])



"""
    GuiSecTools.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy._MOLASS.SerialSettings import get_setting

class GuiSecToolsMenu(Tk.Menu):
    def __init__(self, parent, menubar ):
        self.parent = parent

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        from molass_legacy.Env.EnvInfo import get_global_env_info
        self.env_info = get_global_env_info()
        atsas_state = Tk.NORMAL if self.env_info.atsas_is_available else Tk.DISABLED

        menubar.add_cascade( label="SEC Tools", menu=self )
        # self.add_command( label="Average Subtractor", command=self.show_average_subtractor_dialog )
        self.add_command( label="Elution Curve Picker", command=self.show_elution_curve_picker, state=Tk.DISABLED )
        self.add_command( label="Scattering Curve Plotter", command=self.show_scattering_curve_plotter )
        self.add_command( label="SVD Viewer", command=self.show_svd_viewer, state=Tk.DISABLED )
        self.add_command( label="Average Maker", command=self.show_average_maker)
        # self.add_command( label="Correlation Map Maker", command=self.show_cormap_maker)
        # self.add_command( label="Scaled Data CorMap", command=self.show_scaled_cormap, state=atsas_state)
        # self.add_command( label="Abnormal Data Handling", command=self.show_abnomality_check_dialog, state=Tk.DISABLED )
        # self.add_command( label="Manual Adjuster", command=self.show_manual_adjuster, state=Tk.DISABLED )
        self.add_command( label="Same Sample in Diffrent Conditions", command=self.show_ssdc_analysis)

    def update_states(self):
        use_mtd_conc = get_setting('use_mtd_conc')
        state = Tk.NORMAL if self.parent.dataset_is_ready else Tk.DISABLED

        for k in [0, 2]:
            # if k == 2 and use_mtd_conc:
                # currently disabled due to the error:  AttributeError: 'Absorbance' object has no attribute 'tail_index'
            if k == 2:
                # AttributeError: 'Absorbance' object has no attribute 'baseplane_params'
                state_ = Tk.DISABLED
            else:
                state_ = state
            self.entryconfig(k, state=state_)

    def show_average_subtractor_dialog(self):
        print('show_average_subtractor_dialog')
        from molass_legacy.Microfluidics.AverageSubtractorDialog import AverageSubtractorDialog
        dialog = AverageSubtractorDialog(self.parent, use_mtd=False)
        dialog.show()

    def show_elution_curve_picker(self):
        print('show_data_subtractor_dialog')
        from molass_legacy.Tools.ElutionCurvePicker import ElutionCurvePicker
        parent = self.parent
        dialog = ElutionCurvePicker(parent, parent.pre_recog)
        dialog.show()
        if dialog.applied:
            parent.fig_frame.update_elution_curve()

    def show_scattering_curve_plotter(self):
        print('show_scattering_curve_plotter')
        from molass_legacy.DataStructure.ScatteringCurvePlotter import ScatteringCurvePlotter
        dialog = ScatteringCurvePlotter(self.parent)
        dialog.show()

    def show_svd_viewer(self):
        import Decomposer
        from molass_legacy.DataStructure.SvdViewer import SvdViewer
        from molass_legacy.Mapping.MapperConstructor import create_mapper
        from molass_legacy._MOLASS.SerialSettings import set_setting

        parent = self.parent
        parent.config(cursor='wait')    # does not work. why?
        parent.update()

        analysis_name, analysis_folder = parent.make_analysis_folder()
        set_setting('analysis_folder', analysis_folder)

        pre_recog = parent.pre_recog
        # sd_copy = pre_recog.get_analysis_copy()
        input_sd = parent.serial_data
        sd_copy = input_sd._get_analysis_copy_impl(pre_recog)
        in_folder = self.parent.in_folder.get()
        title = "SVD Viewer - " + in_folder
        mapper = create_mapper( parent, sd_copy, input_sd, pre_recog, analyzer_dialog=parent, logger=parent.tmp_logger )
        viewer = SvdViewer( parent, title, sd_copy, mapper )

        parent.config(cursor='')

        viewer.show()

    def show_average_maker(self):
        from SecTools.AverageMakerDialog import AverageMakerDialog
        dialog = AverageMakerDialog(self.parent)
        dialog.show()

    def show_cormap_maker(self):
        from SecTools.CorMap.CormapMakerDialog import CormapMakerDialog
        atsas = self.env_info.atsas_is_available
        dialog = CormapMakerDialog(self.parent, atsas=atsas)
        dialog.show()

    def show_scaled_cormap(self):
        from SecTools.CorMap.ScaledDataDialog import ScaledDataDialog
        atsas = self.env_info.atsas_is_available
        dialog = ScaledDataDialog(self.parent, atsas=atsas)
        dialog.show()

    def show_abnomality_check_dialog(self):
        print('show_abnomality_check_dialog')

    def show_manual_adjuster(self):
        from molass_legacy.Mapping.ManualAdjuster import ManualAdjuster
        parent = self.parent
        sd = parent.pre_recog.pre_recog_copy
        ms = ManualAdjuster( parent, sd )
        ms.show()
        if ms.applied:
            # TODO:
            pass

    def show_ssdc_analysis(self, devel=True):
        from importlib import reload
        if devel:
            import SSDC.FolderEntryDialog
            reload(SSDC.FolderEntryDialog)
        from SSDC.FolderEntryDialog import FolderEntryDialog
        dialog = FolderEntryDialog(self.parent)
        dialog.show()
        if dialog.applied:
            if devel:
                import SSDC.SsdcAnalysis
                reload(SSDC.SsdcAnalysis)
            from SSDC.SsdcAnalysis import SsdcAnalysis
            analysis = SsdcAnalysis(self.parent, dialog.get_info_list())
            analysis.show()



# coding: utf-8
"""
    SecTheory.FitTrialNonCf.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Batch.StandardProcedure import StandardProcedure
from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.Peaks.ElutionModels import egh
from LPM import get_corrected
from DataUtils import get_in_folder
from SecSaxs.DataTreatment import DataTreatment

class DemoDialog(Dialog):
    def __init__(self, parent, ecurve, rg_curve, rp_params, rg_params, wt_params):
        self.ecurve = ecurve
        self.rg_curve = rg_curve
        self.rp_params = rp_params
        self.rg_params = rg_params
        self.wt_params = wt_params
        Dialog.__init__(self, parent, "Fitting Test", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        left_frame = Tk.Frame(body_frame)
        left_frame.grid(row=0, column=0)
        right_frame = Tk.Frame(body_frame)
        right_frame.grid(row=0, column=1)

        # 
        self.coding_text = Tk.Text(left_frame)
        self.coding_text.pack()
        for varname, params in [("rp_params", self.rp_params),
                                ("rg_params", self.rg_params),
                                ("wt_params", self.wt_params),
                                ]:
            self.coding_text.insert(Tk.END, varname + " = " + str(list(params)) + "\n")

        # 
        row = 0

        self.n_species = Tk.IntVar()
        self.n_species.set(len(self.rg_params))

        label = Tk.Label(right_frame, text="n_species: ")
        label.grid(row=row, column=0, sticky=Tk.E, pady=10)

        sbox  = Tk.Spinbox(right_frame, textvariable=self.n_species,
                          from_=0, to=30, increment=1,
                          justify=Tk.CENTER, width=6, state=Tk.DISABLED)
        sbox.grid(row=row, column=1, sticky=Tk.W, pady=10)

        row += 1

        self.n_iters = Tk.IntVar()
        self.n_iters.set(10)

        label = Tk.Label(right_frame, text="n_iters: ")
        label.grid(row=row, column=0, sticky=Tk.E, pady=10)

        sbox  = Tk.Spinbox(right_frame, textvariable=self.n_iters,
                          from_=0, to=100, increment=1,
                          justify=Tk.CENTER, width=6)
        sbox.grid(row=row, column=1, sticky=Tk.W, pady=10)

        row += 1

        self.n_trials = Tk.IntVar()

        label = Tk.Label(right_frame, text="n_trials: ")
        label.grid(row=row, column=0, sticky=Tk.E, pady=10)

        sbox  = Tk.Spinbox(right_frame, textvariable=self.n_trials,
                          from_=0, to=30, increment=1,
                          justify=Tk.CENTER, width=6)
        sbox.grid(row=row, column=1, sticky=Tk.W, pady=10)

        row += 1

        run_btn = Tk.Button(right_frame, text="Run", command=self.run)
        run_btn.grid(row=row, column=0, columnspan=2, pady=10)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.bind("<Escape>", self.cancel)

    def run(self):
        from importlib import reload
        import SecTheory.SinglePoreNonCf
        reload(SecTheory.SinglePoreNonCf)
        from .SinglePoreNonCf import fit_single_pore_pdf
        text = self.coding_text.get("1.0", Tk.END)
        temp_namespace = {}
        exec(text, temp_namespace)
        params_list = []
        for varname in ["rp_params", "rg_params", "wt_params"]:
            params_list.append(np.array(temp_namespace.get(varname)))

        init_params = np.concatenate(params_list)
        fit_single_pore_pdf(self.ecurve, self.rg_curve, init_params,
                                n_species=len(params_list[1]),
                                n_iters=self.n_iters.get(),
                                n_trials=self.n_trials.get(),
                                )

def fitting_demo(root, in_folder, rg_folder, rp_params, rg_params, wt_params):

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=2, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()

    if rg_folder is None:
        assert False
    else:
        from RgProcess.RgCurveProxy import RgCurveProxy
        rg_curve = RgCurveProxy(ecurve, rg_folder)

    dialog = DemoDialog(root, ecurve, rg_curve, rp_params, rg_params, wt_params)
    dialog.geometry("+500+500")
    dialog.show()



"""
    Batch.LiteBatch.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import logging
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Optimizer.FuncImporter import get_objective_function_info
from .OptDataSetsProxy import OptDataSetsProxy as OptDataSets
from .FullBatch import FullBatch

class LiteBatch(FullBatch):
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.unified_baseline_type = get_setting("unified_baseline_type")

    def get_init_estimate(self):
        self.func_info = get_objective_function_info(self.logger, default_func_code='G0346')
        self.func_dict = self.func_info.func_dict
        self.key_list = self.func_info.key_list
        self.dsets = OptDataSets(self.sd, self.corrected_sd)
        self.construct_optimizer()
        init_params = self.compute_init_params()
        return init_params

def create_corrected_sd(sd):
    lb = LiteBatch()
    lb.prepare(sd)
    return lb.corrected_sd



# -*- coding: utf-8 -*-
"""

    ファイル名：   OurException.py

    処理内容：

        例外処理

"""
from __future__ import division, print_function, unicode_literals

class OurException( Exception ):
    def __init__( self, value ):
        self.value   = value

    def __str__( self ):
        return repr( self.value )



# -*- coding: utf-8 -*-
"""

    ファイル名：   DebugQueue.py

    処理内容：

        テスト用の情報を伝達する。

        cf. http://stackoverflow.com/questions/7109093/checking-for-empty-queue-in-pythons-multiprocessing

"""
from __future__ import division, print_function, unicode_literals

import sys
import queue
import logging

debug_queue = queue.Queue()

def debug_queue_put( data ):
    debug_queue.put( data )

def debug_queue_get():
    try:
        return debug_queue.get( block = False )
    except queue.Empty:
        return None
    except:
        logging.exception( 'Unexpected error' )
        raise



"""
    SecTheory.NumPlates.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Elution.CurveUtils import simple_plot
from .RetensionTime import compute_retention_time
from .MonoPore import compute_mu_sigma

NUM_PLATES = 48000*0.3

def contrast_sigma_info(in_folder, sd, job, nc, fv, seccol_params, sigmas, monopore_params, rg_params, debug=True):
    xr_curve = sd.get_xray_curve()

    emg_peaks = xr_curve.get_emg_peaks()
    p = xr_curve.primary_peak_no
    ppeak = emg_peaks[p]
    x = xr_curve.x
    params = ppeak.get_params()
    print("params=", params)
    mean = params[1]
    stdev = params[2]

    # get the nearest Rg
    trs = compute_retention_time(seccol_params, rg_params)
    n = np.argmin((trs - mean)**2)
    rg = rg_params[n]
    if sigmas is None:
        # stc
        mu, sigma = compute_mu_sigma(*monopore_params, rg)
    else:
        # egh
        mu = mean
        sigma = sigmas[n]

    tr = compute_retention_time_from_N(NUM_PLATES, sigma)
    print("tr=", tr)

    t0, K, rp, m = seccol_params

    if debug:
        from DataUtils import get_in_folder
        in_folder = get_in_folder(in_folder)
        with plt.Dp():
            fig, ax = plt.subplots(figsize=(18,4))
            ax.set_title("SEC Parameter Plot at Job %s processing Data %s" % (job, in_folder), fontsize=16)
            simple_plot(ax, xr_curve, legend=False)
            y = ppeak.get_model_y(x)
            ax.plot(x, y, label="egh")
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            ti = params[1] - tr
            for j, color, label in (ti,"green", "injection"), (t0, "yellow", "t0"), (t0+K, "yellow", "t0+K"):
                ax.plot([j, j], [ymin, ymax], color=color, label=label)

            xmin, xmax = ax.get_xlim()
            dx = (xmax - xmin)*0.02
            dy = (ymax - ymin)*0.05

            px = mean
            py = xr_curve.spline(px)
            ax.text(px-dx, py-dy, "Rg = %.3g" % rg, ha="right", va="center", fontsize=20, alpha=0.5)

            tx = xmin*0.5 + xmax*0.5
            ty = ymin*0.2 + ymax*0.8
            ax.text(tx, ty, "Pore Size = %.3g" % rp, ha="center", va="center", fontsize=20, alpha=0.5)
            ax.legend(loc="lower center")
            fig.tight_layout()
            plt.show()

    return mean, stdev, mu, sigma, tr, rg

def compute_retention_time_from_N(num_plates, sigma):
    """
    N = (tr/stdev)**2
    tr = sqrt(N)*sigma
    """
    return np.sqrt(num_plates) * sigma



"""
    HdcTheory.ElutionCurve.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

NUM_PLATES_PC = 14400
C = 2.698   # 

def retention_time(tp, reff, r0, C):
    lamb = reff/r0
    return tp / (1 + 2*lamb - C*lamb**2)

def compute_elution_curve(x, t0, reff, r0, sigma, Npc=NUM_PLATES_PC, C=C):
    from molass_legacy.Peaks.ElutionModels import gaussian
    tp = np.sqrt(Npc)*sigma
    tI = t0 - tp
    tR = tI + retention_time(tp, reff, r0, C)
    print("tR=", tR)
    return gaussian(x, 1, tR, sigma)

def demo():
    import matplotlib.pyplot as plt
    import seaborn as sns
    sns.set_theme()
    x = np.arange(500)
    t0 = 200
    sigma = 10
    r0 = 5000
    tp = np.sqrt(NUM_PLATES_PC)*sigma
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    for rg in [200, 150, 75, 20]:
        tR = retention_time(tp, rg, r0, C)
        ax1.plot(tR, rg, "o", label=r'$R_g$=%g' % rg)
        y = compute_elution_curve(x, t0, rg, r0, 10)
        ax2.plot(x, y, label=r'$R_g$=%g' % rg)
    ax2.axvline(x=t0, color='red', label=r"$t_0$")
    ax1.legend()
    ax2.legend()
    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    # import seaborn as sns
    # sns.set_theme()
    sys.path.append("../lib")
    demo()



"""
    RgSecUpdater.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from SecTheory.SecEstimator import guess_initial_secparams

USE_BASIN_HOPPING = True
if USE_BASIN_HOPPING:
    from scipy.optimize import basinhopping
else:
    from scipy.optimize import minimize

def update_optimizer_impl(optimizer, params, rg_params, rg_qualities):
    print("update_optimizer_impl: rg_params=", rg_params, "rg_qualities=", rg_qualities)

    separate_params = optimizer.split_params_simple(params)
    separate_params[2] = rg_params

    optimizer.init_rgs = rg_params

    init_xr_params = separate_params[0]
    Npc, rp, tI, t0, P, m = guess_initial_secparams(init_xr_params, rg_params)
    trs = init_xr_params[:,1]
    mu, sigma = init_xr_params[0,[1,2]]

    bounds = [(tI, mu-sigma), (0, P*2), (0, 3)]
    init_params = [t0, P, m]
    rhos = rg_params/rp
    rhos[rhos > 1] = 1

    def fixed_poresize_objective(p):
        t0_, P_, m_ = p
        trs_ = t0_ + P_*np.power(1 - rhos, m_)
        return np.sum(((trs_ - trs)*rg_qualities)**2)

    objective_func = fixed_poresize_objective

    if USE_BASIN_HOPPING:
        minimizer_kwargs = dict(method="Nelder-Mead", bounds=bounds)
        ret = basinhopping(objective_func, init_params, minimizer_kwargs=minimizer_kwargs)
    else:
        ret = minimize(objective_func, init_params, method="Nelder-Mead", bounds=bounds)

    rp_ = rp
    t0_, P_, m_ = ret.x

    init_sec_params = np.array([Npc, rp_, tI, t0_, P_, m_])
    separate_params[7] = init_sec_params

    optimizer.init_separate_params = separate_params

    x = np.concatenate([init_xr_params.flatten(),*separate_params[1:]])
    optimizer.init_params = x
    optimizer.update_bounds(x)

    return x
